# Copyright 2018 Olivier Scholder <o.scholder@gmail.com>

"""
Library to handle SPM data.
This is the core module of all images retrieved by SPM and ToF-SIMS.
"""

import copy

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import scipy
import scipy.interpolate
import scipy.ndimage
import scipy.optimize
import skimage
import skimage.exposure
import skimage.filters
from skimage import transform as tf
from skimage.draw import disk

from .utils import CDF, funit
from .utils.misc import PB

try:
    from skimage.filters import threshold_local
except:
    # For compatibility with old versions of skimage
    from skimage.filters import threshold_adaptive as threshold_local


class SPM_image:
    """
    Main class to handle SPM images.
    This class contains the pixels data of the images as well as it's real size.
    It also provides a lot of tools to correct and perform various analysis and tasks on the image.
    """

    def __init__(
        self,
        BIN,
        channel="Topography",
        corr=None,
        real=None,
        zscale="?",
        _type="Unknown",
    ):
        """
        Create a new SPM_image

        Parameters
        ----------
        BIN : 2D numpy array
            The pixel values of the image as a 2D numpy array
        channel : string
            The name of the channel. What does the image represents?
        corr : string or None
            'slope' : correct the SPM image for its slope (see pySPM.SPM.SPM_image.correct_slope)
            'lines' : correct the SPM image for its lines (see pySPM.SPM.SPM_image.correct_lines)
            'plane' : correct the SPM image by plane fitting (see pySPM.SPM.SPM_image.correct_plane)
        real : None or dictionary
            Information about the real size of the image {'x':width,'y':height,'unit':unit_name}
        zscale : string
            Unit used to describe the z-scale. (units of the data of BIN)
        _type : string
            represent the type of measurement
        """
        self.channel = channel
        self.direction = "Unknown"
        self.size = {"pixels": {"x": BIN.shape[1], "y": BIN.shape[0]}}
        if real is not None:
            self.size["real"] = real
        else:
            self.size["real"] = {"unit": "pixels", "x": BIN.shape[1], "y": BIN.shape[0]}
        if "unit" not in self.size["real"]:
            self.size["real"]["unit"] = "px"
        self.pixels = BIN
        self.type = _type
        self.zscale = zscale
        if corr is not None:
            if corr.lower() == "slope":
                self.correct_slope()
            elif corr.lower() == "lines":
                self.correct_lines()
            elif corr.lower() == "plane":
                self.correct_plane()

    def __add__(self, b):
        """
        Add up two images. This is a low level function and no check is performed to proof that both images have the same size.
        """
        New = copy.deepcopy(self)
        if isinstance(b, SPM_image):
            New.pixels += b.pixels
            New.channel += " + " + b.channel
        elif type(b) in [int, float]:
            New.pixels += b
            New.channels += f" + {b:.2f}"
        return New

    def __sub__(self, b):
        """
        Subtract two images. This is a low level function and no check is performed to proof that both images have the same size.
        """
        New = copy.deepcopy(self)
        if isinstance(b, SPM_image):
            New.pixels -= b.pixels
            New.channel += " - " + b.channel
        elif type(b) in [int, float]:
            New.pixels -= b
            New.channels += f" - {b:.2f}"
        return New

    def __mul__(self, b):
        New = copy.deepcopy(self)
        if isinstance(b, SPM_image):
            New.pixels *= b.pixels
            New.channel = f"({New.channel})*{b.channel}"
        elif type(b) in [int, float]:
            New.pixels *= b
            New.channels = f"({New.channel})*{b:.2f}"
        return New

    def __div__(self, b):
        New = copy.deepcopy(self)
        if isinstance(b, SPM_image):
            New.pixels /= b.pixels
            New.channel = f"({New.channel})/{b.channel}"
        elif type(b) in [int, float]:
            New.pixels /= b
            New.channels = f"({New.channel})/{b:.2f}"
        return New

    def pxs(self):
        """
        Return the pixel size
        """
        fxy = {
            xy: funit(self.size["real"][xy], self.size["real"]["unit"]) for xy in "xy"
        }
        return [
            (fxy[xy]["value"] / self.size["pixels"][xy], fxy[xy]["unit"]) for xy in "xy"
        ]

    def add_scale(
        self,
        length,
        ax=None,
        height=20,
        margin=5,
        color="w",
        loc=4,
        text=True,
        pixels=None,
        fontsize=20,
        edge_color="k",
        edge_width=3,
    ):
        """
        Display a scale marker on an existing image

        Parameters
        ----------

        length : float
            The length of the scale in real units
        ax : matplotlib axis
            if None the current axis will be taken (plt.gca())
        height : int
            The height of the scale bar in pixels
        color : string
            The color used to display the scale bar
        loc : int
            The location of the scale bar.
            1 : top right
            2 : top left
            3 : bottom left
            4 : bottom right
        text : bool
            display the size of the scale on top of it?
        pixels : bool
            Is the image plotted in ax with a x/y scale in pixels?
        fontsize : float
            The fontsize used to display the text

        Example
        -------
        >>> img = pySPM.SPM_image()
        >>> img.show()
        >>> img.add_scale(50e-6, pixels=False);
        Add a scale of 50 μm on an image displayed with real units

        >>> img = pySPM.SPM_image()
        >>> img.show(pixels=True)
        >>> img.add_scale(50e-6);
        Add a scale of 50 μm on an image displayed in pixels
        """

        import matplotlib.patches
        import matplotlib.patheffects

        fL = length / self.size["real"]["x"]
        self.size["pixels"]["x"] * fL
        fH = height / self.size["pixels"]["y"]
        if ax is None:
            ax = plt.gca()
        if pixels is None:
            pixels = ax.isPixel if hasattr(ax, "isPixel") else False
        if hasattr(ax, "flipped"):
            pass
        if isinstance(loc, int):
            assert loc in [1, 2, 3, 4]
            ref = ax.transAxes.transform(
                {1: (1 - fL, 0), 2: (0, 0), 3: (0, 1 - fH), 4: (1 - fL, 1 - fH)}[loc]
            )
            if loc in [2, 3]:
                ref[0] += margin
            else:
                ref[0] -= margin
            if loc in [1, 2]:
                ref[1] += margin
            else:
                ref[1] -= margin
        else:
            assert type(loc) in [tuple, list]
            assert len(loc) == 2
            ref = (
                ax.transData.transform(loc)
                + ax.transAxes.transform((-fL / 2, -fH / 2))
                - ax.transAxes.transform((0, 0))
            )
        inv = ax.transData.inverted()
        ref = inv.transform(ref)
        WH = inv.transform(ax.transAxes.transform((fL, fH))) - inv.transform(
            ax.transAxes.transform((0, 0))
        )
        ax.add_patch(
            matplotlib.patches.Rectangle(ref, width=WH[0], height=WH[1], color=color)
        )

        if text:
            r = funit(length, self.size["real"]["unit"])
            if r["unit"][0] == "u":
                r["unit"] = "µ" + r["unit"][1:]
            if loc in [3, 4]:
                label_ref = [ref[0] + WH[0] / 2, ref[1]]
                ann = ax.annotate(
                    "{value:.01f} {unit}".format(**r),
                    label_ref,
                    color=color,
                    fontsize=fontsize,
                    va="top",
                    ha="center",
                )
            else:
                label_ref = [ref[0] + WH[0] / 2, ref[1] + WH[1]]
                ann = ax.annotate(
                    "{value:.01f} {unit}".format(**r),
                    label_ref,
                    color=color,
                    fontsize=fontsize,
                    va="bottom",
                    ha="center",
                )
            ann.set_path_effects(
                [
                    matplotlib.patheffects.withStroke(
                        linewidth=edge_width, foreground=edge_color
                    )
                ]
            )

    def offset(self, profiles, width=1, ax=None, col="w", inline=True, **kargs):
        """
        Correct an image by offsetting each row individually in order that the lines passed as argument in "profiles" becomes flat.

        Parameters
        ----------
        profiles: list of list
            each sublist represent a line as [x1, y1, x2, y2] in pixels known to be flat
        width : int, float
            the line width in pixels used for better statistics
        ax : matplotlib axis or None
            If not None, axis in which the profiles will be plotted in
        inline : bool
            If True perform the correction on the current object, otherwise return a new image
        col : string
            matrplotlib color used to plot the profiles (if ax is not None)
        labels : bool
            display a label number with each profile
        **kargs: arguments passed further to get_row_profile.
            axPixels: set to True if you axis "ax" have the data plotted in pixel instead of real distance

        Example
        -------
        Exampel if the data are plotted in pixels:
        >>> topo = pySPM.SPM_image(...)
        >>> fig, ax = plt.subplots(1, 2, figsize=(10, 5))
        >>> topoC = topo.offset([[150, 0, 220, 255]], inline=False,axPixels=True)
        >>> topo.show(pixels=True, ax=ax[0])
        >>> topoC.show(ax=ax[1]);

        Example if the data are plotted with real units
        >>> topo = pySPM.SPM_image(...)
        >>> fig, ax = plt.subplots(1, 2, figsize=(10, 5))
        >>> topoC = topo.offset([[150, 0, 220, 255]], inline=False)
        >>> topo.show(ax=ax[0])
        >>> topoC.show(ax=ax[1]);
        """
        offset = np.zeros(self.pixels.shape[0])
        counts = np.zeros(self.pixels.shape[0])
        for i, p in enumerate(profiles):
            if kargs.get("labels", False):
                y, D = self.get_row_profile(
                    *p, width=width, ax=ax, col=col, label=str(i), **kargs
                )
            else:
                y, D = self.get_row_profile(*p, width=width, ax=ax, col=col, **kargs)
            counts[y] += 1
            offset[y[1:]] += np.diff(D)
        counts[counts == 0] = 1
        offset = offset / counts
        offset = np.cumsum(offset)
        offset = offset.reshape((self.pixels.shape[0], 1))
        if inline:
            self.pixels = self.pixels - np.flipud(
                np.repeat(offset, self.pixels.shape[1], axis=1)
            )
            return self
        else:
            C = copy.deepcopy(self)
            C.pixels = self.pixels - np.flipud(
                np.repeat(offset, self.pixels.shape[1], axis=1)
            )
            return C

    def pxRect2Real(self, xy, width, height):
        """
        Transform a xy, width, height data in pixels to an equivalentz one with real units
        """
        ll = self.px2real(xy[0], xy[1])
        ur = self.px2real(xy[0] + width, xy[1] + height)
        return ll, ur[0] - ll[0], ur[1] - ll[1]

    def get_row_profile(
        self, x1, y1, x2, y2, width=1, col="C1", ax=None, alpha=0, **kargs
    ):
        """
        Get a profile per row along a given line. This function is mainly useful for the function offset.

        x1, y1, x2, y2: int
            coordinates of the line.
        width : int
            the width of the line used for statistics (in pixels)
        col: string
            color used to plot the line position
        ax : matplotlib axis
            axis in which the lines position will plotted
        alpha : float
            The alpha channel of the line color (≥0 and ≤1)
        **kargs:
            line style arguments: linewidth, color and linestyle
            axis units: axPixels set to True if ax has the image plotted in pixels.

        Returns
        -------
        Y coordinates : 1D numpy array
            distance along the profile starting at 0
        Z coordinates : 1D numpy array
            profile
        """
        plotargs = {
            key: kargs[key]
            for key in ["linewidth", "color", "linestyle"]
            if key in kargs
        }
        if y2 < y1:
            x1, y1, x2, y2 = x2, y2, x1, y1
        if ax is not None:
            d = np.sqrt((x2 - x1) ** 2 + (y2 - y1) ** 2)
            dx = -width / 2 * (y2 - y1) / d
            dy = width / 2 * (x2 - x1) / d
            if kargs.get("axPixels", False):
                ax.plot([x1 - dx, x1 + dx], [y1 - dy, y1 + dy], col)
                ax.plot([x2 - dx, x2 + dx], [y2 - dy, y2 + dy], col)
                ax.plot((x1, x2), (y1, y2), col, **plotargs)
                if kargs.get("label", False):
                    ax.annotate(
                        kargs.get("label"),
                        (0.5 * (x1 + x2), 0.5 * (y1 + y2)),
                        color=col,
                    )
                if alpha > 0:
                    import matplotlib.patches

                    ax.add_patch(
                        matplotlib.patches.Rectangle(
                            (x1 + dx, y1 + dy),
                            width,
                            d,
                            -np.degrees(np.arctan2(x2 - x1, y2 - y1)),
                            color=col,
                            alpha=alpha,
                        )
                    )
            else:
                h = self.pixels.shape[0]
                pxs = self.size["real"]["x"] / self.pixels.shape[1]
                pys = self.size["real"]["y"] / h
                ax.plot(
                    [(x1 - dx) * pxs, (x1 + dx) * pxs],
                    [(h - (y1 - dy)) * pys, (h - (y1 + dy)) * pys],
                    col,
                )
                ax.plot(
                    [(x2 - dx) * pxs, (x2 + dx) * pxs],
                    [(h - (y2 - dy)) * pys, (h - (y2 + dy)) * pys],
                    col,
                )
                ax.plot(
                    (x1 * pxs, x2 * pxs),
                    ((h - y1) * pys, (h - y2) * pys),
                    col,
                    **plotargs,
                )
                if kargs.get("label", False):
                    ax.annotate(
                        kargs.get("label"),
                        (0.5 * (x1 + x2) * pxs, 0.5 * (2 * h - y1 - y2) * pys),
                        color=col,
                    )
                if alpha > 0:
                    import matplotlib.patches

                    W = np.sqrt((2 * dx * pxs) ** 2 + (2 * dy * pys) ** 2)
                    L = np.sqrt(((x2 - x1) * pxs) ** 2 + ((y2 - y1) * pys) ** 2)
                    ax.add_patch(
                        matplotlib.patches.Rectangle(
                            ((x1 + dx) * pxs, (y1 + dy) * pys),
                            W,
                            L,
                            -np.degrees(np.arctan2((x2 - x1) * pxs, (y2 - y1) * pys)),
                            color=col,
                            alpha=alpha,
                        )
                    )

        x = np.arange(self.pixels.shape[1])
        y = np.arange(self.pixels.shape[0])
        I = scipy.interpolate.interp2d(x, y, np.flipud(self.pixels))

        Y = np.arange(y1, y2 + 1)
        V = np.zeros(len(Y))
        for w in np.arange(width):
            xl = np.linspace(
                x1 - (width - 1) / 2.0 + w, x2 - (width - 1) / 2.0 + w, len(Y)
            )
            for i in range(len(Y)):
                Z = I(xl[i], Y[i])
                V[i] += Z
        return Y, V / width

    def correct_median_diff(self, inline=True):
        """
        Correct the image with the median difference
        """
        N = self.pixels
        # Difference of the pixel between two consecutive row
        N2 = N - np.vstack([N[:1, :], N[:-1, :]])
        # Take the median of the difference and cumsum them
        C = np.cumsum(np.median(N2, axis=1))
        # Extend the vector to a matrix (row copy)
        D = np.tile(C, (N.shape[0], 1)).T
        if inline:
            self.pixels = N - D
        else:
            New = copy.deepcopy(self)
            New.pixels = N - D
            return New

    def correct_slope(self, inline=True):
        """
        Correct the image by subtracting a fitted slope along the y-axis
        """
        s = np.mean(self.pixels, axis=1)
        i = np.arange(s.shape[0])
        fit = np.polyfit(i, s, 1)
        if inline:
            self.pixels -= np.tile(
                np.polyval(fit, i).reshape(self.pixels.shape[0], 1),
                self.pixels.shape[1],
            )
            return self
        else:
            New = copy.deepcopy(self)
            New.pixels -= np.tile(
                np.polyval(fit, i).reshape(self.pixels.shape[0], 1),
                self.pixels.shape[1],
            )
            return New

    def correct_plane(self, inline=True, mask=None):
        """
        Correct the image by subtracting a fitted 2D-plane on the data

        Parameters
        ----------
        inline : bool
            If True the data of the current image will be updated otherwise a new image is created
        mask : None or 2D numpy array
            If not None define on which pixels the data should be taken.
        """
        x = np.arange(self.pixels.shape[1])
        y = np.arange(self.pixels.shape[0])
        X0, Y0 = np.meshgrid(x, y)
        Z0 = self.pixels
        if mask is not None:
            X = X0[mask]
            Y = Y0[mask]
            Z = Z0[mask]
        else:
            X = X0
            Y = Y0
            Z = Z0
        A = np.column_stack((np.ones(Z.ravel().size), X.ravel(), Y.ravel()))
        c, resid, rank, sigma = np.linalg.lstsq(A, Z.ravel(), rcond=-1)
        if inline:
            self.pixels -= c[0] * np.ones(self.pixels.shape) + c[1] * X0 + c[2] * Y0
            return self
        else:
            New = copy.deepcopy(self)
            New.pixels -= c[0] * np.ones(self.pixels.shape) + c[1] * X0 + c[2] * Y0
            return New

    def correct_lines(self, inline=True):
        """
        Subtract the average of each line for the image.

        if inline is True the current data are updated otherwise a new image with the corrected data is returned
        """
        if inline:
            self.pixels -= np.tile(
                np.mean(self.pixels, axis=1).T, (self.pixels.shape[1], 1)
            ).T
            return self
        else:
            New = copy.deepcopy(self)
            New.pixels -= np.tile(
                np.mean(self.pixels, axis=1).T, (self.pixels.shape[1], 1)
            ).T
            return New

    def dist_v2(self, pixel=False):
        """
        Return a 2D array with the distance between each pixel and the closest border.
        Might be usefull for FFT filtering
        """
        if pixel:
            dx = 1
            dy = 1
        else:
            dx = self.size["real"]["x"] / self.size["pixels"]["x"]
            dy = self.size["real"]["y"] / self.size["pixels"]["y"]
        x2 = np.arange(self.size["pixels"]["x"])
        x2 = (np.minimum(x2, self.size["pixels"]["x"] - x2) * dx) ** 2
        y2 = np.arange(self.size["pixels"]["y"])
        y2 = (np.minimum(y2, self.size["pixels"]["y"] - y2) * dy) ** 2
        X, Y = np.meshgrid(x2, y2)
        return np.sqrt(X + Y)

    def inv_calc_flat(self, d, l=0.1):
        """
        Function used for inverse MFM calculation (inspired from http://qmfm.empa.ch/qmfm/)
        The function is in its early devlopment stage as not used by the developed.

        Parameters
        ----------
        d : float
            Height distance in the input data
        l : float
            Tikhonov parameter for the deconvolution
        """
        work_image = self.pixels
        ny, nx = self.pixels.shape
        self.size["real"]["x"] / self.size["pixels"]["x"]
        self.size["real"]["y"] / self.size["pixels"]["y"]

        k = self.dist_v2()
        k[0, 0] = 1e-10

        tf = np.exp(-d * k)
        tf[0, 0] = np.mean(tf)
        tf /= 2
        tf *= 1 - np.exp(-d * k)

        recon_tf = np.ones(tf.shape) / (tf + l * np.ones(tf.shape) / np.conj(tf))
        tf *= recon_tf
        return np.real(np.fft.ifft2(np.fft.fft2(work_image) * recon_tf))

    def get_extent(self):
        """
        Get the image extent in real data
        """
        if "recorded" in self.size:
            W = self.size["recorded"]["real"]["x"]
            H = self.size["recorded"]["real"]["y"]
        else:
            W = self.size["real"]["x"]
            H = self.size["real"]["y"]
        return (0, W, 0, H)

    def show(
        self,
        ax=None,
        sig=None,
        cmap=None,
        title=None,
        adaptive=False,
        dmin=0,
        dmax=0,
        pixels=False,
        flip=False,
        wrap=None,
        mul=1,
        symmetric=False,
        **kargs,
    ):
        """
        Function to display the image with a lot of parametrization

        Parameters
        ----------
        ax : matplotlib axis or None
            matplotlib axis if given otherwise current axis will be used (plt.gca())
        sig : float
            sigma values to adjust the contrast range around the mean ±sig times the standard-deviation
        cmap : string
            colormap name used. By default a gray map is used. If the zscale of the data are in 'meter' (i.e. topography data) the 'hot' colormap is used
        title : string
            The title of the plot. By default is the channel name
        adaptive : bool
            The color scale used is linear. If adaptive is True a non linear color scale is used in order that each color is used with the same amount.
        dmin : float
            minimum value adjustment used for the colorscale
        dmax: float
            maximum value adjustment used for the colorscale
        pixels : bool
            Display the image with x/y-labels with real unit. If pixels is True, the axes are in pixels
        flip : bool
            Flip the image upside-down
        wrap : Nont or int
            wrap the title to a width of wrap chars
        symmetric : bool
            If True will place the middle of the colorscale to the value 0.
            This is specially usefull for diverging colormaps such as : BrBG, bwr, coolwarm, seismiv, spectral, etc.
        level : float
            level should be ≥0 and <50. Adjust the lower and upper colorscale to level% and (100-level)% of the data range.
            e.g. if level=1, the colorscale will display 1-99% of the data range
        vmin : float
            Minimum value used for the colorscale
        vmax : flaot
            Maximum value used for the colorscale


        Returns
        -------
        matplotlib.image.AxesImage
            matplolib axis instance returned by imshow

        Examples
        --------
        >>> topo = pySPM.SPM_image(...)
        >>> fig, (ax, ax2) = plt.subplots(2, 3, figsize=(15, 10))
        >>> topo.show(ax=ax[0], cmap='gray', title="color map=\"gray\"")
        >>> topo.show(ax=ax[1], sig=2, title="standard deviation=2")
        >>> topo.show(ax=ax[2], adaptive=True, title="Adaptive colormap")
        >>> topo.show(ax=ax2[0], dmin=4e-8, cmap='gray', title="raise the lowest value for the colormap of +40nm")
        >>> topo.show(ax=ax2[1], dmin=3e-8, dmax=-3e-8, cmap='gray',title="raise lower of +30nm and highest of -30nm")
        >>> topo.show(ax=ax2[2], pixels=True, title="Set axis value in pixels");
        """
        mpl.rc("axes", grid=False)

        if ax is None:
            ax = plt.gca()
        ax.src = self
        if title is None:
            title = f"{self.type} - {self.channel}"
        if wrap is not None:
            title = "\n".join(
                [
                    title[i * wrap : (i + 1) * wrap]
                    for i in range(int(len(title) / wrap) + 1)
                ]
            )
        unit = self.size["real"]["unit"]
        sunit = "afpnum kMGTPE"
        if len(unit) == 1 or unit in ["pixels"]:
            isunit = 6
        elif unit[0] in sunit:
            isunit = sunit.find(unit[0])
            unit = unit[1:]
        else:
            isunit = 6
        W = self.size["real"]["x"]
        H = self.size["real"]["y"]
        fact = int(np.floor(np.log(W) / np.log(10) / 3))
        isunit += fact
        W, H = W / 10 ** (fact * 3), H / 10 ** (fact * 3)
        if cmap is None:
            cmap = "gray"
            if unit == "m" and self.channel == "Topography":
                cmap = "hot"
        mi, ma = np.nanmin(self.pixels), np.nanmax(self.pixels)
        if adaptive:
            img = np.asarray(256**2 * (self.pixels - mi) / (ma - mi), dtype=np.uint16)
            mi, ma = 0, 1
            img = skimage.exposure.equalize_adapthist(img, clip_limit=0.03)
        else:
            img = mul * self.pixels
            mi *= mul
            ma *= mul

        if sig is None:
            vmin = mi + dmin
            vmax = ma + dmax
        else:
            std = np.nanstd(img)
            avg = np.nanmean(img)
            vmin = avg - sig * std
            vmax = avg + sig * std
        if "level" in kargs:
            if kargs["level"] < 0 or kargs["level"] >= 50:
                raise ValueError("The level shoud have a value in [0,50)")
            vmax = np.percentile(img, 100 - kargs["level"])
            vmin = np.percentile(img, kargs["level"])
            del kargs["level"]
        if "vmin" in kargs:
            vmin = kargs["vmin"]
            del kargs["vmin"]
        if "vmax" in kargs:
            vmax = kargs["vmax"]
            del kargs["vmax"]
        if symmetric:
            vmax = abs(max(vmin, vmax))
            vmin = -vmax
        if not flip:
            ax.flipped = False
            if pixels:
                ax.isPixel = True
                r = ax.imshow(
                    np.flipud(img),
                    extent=[0, img.shape[1], img.shape[0], 0],
                    cmap=cmap,
                    vmin=vmin,
                    vmax=vmax,
                    **kargs,
                )
            else:
                ax.isPixel = False
                r = ax.imshow(
                    np.flipud(img),
                    extent=[0, W, 0, H],
                    cmap=cmap,
                    vmin=vmin,
                    vmax=vmax,
                    **kargs,
                )
        else:
            ax.flipped = True
            if pixels:
                ax.isPixel = True
                r = ax.imshow(
                    np.flipud(img),
                    extent=[0, img.shape[1], img.shape[0], 0],
                    cmap=cmap,
                    vmin=vmin,
                    vmax=vmax,
                    **kargs,
                )
            else:
                ax.isPixel = False
                r = ax.imshow(
                    np.flipud(img),
                    cmap=cmap,
                    extent=[0, W, 0, H],
                    vmin=vmin,
                    vmax=vmax,
                    **kargs,
                )
        if pixels:
            ax.set_xlim((0, self.pixels.shape[1]))
            if flip:
                ax.set_ylim((0, self.pixels.shape[0]))
            else:
                ax.set_ylim((self.pixels.shape[0], 0))
        else:
            ax.set_xlim((0, W))
            if flip:
                ax.set_ylim((H, 0))
            else:
                ax.set_ylim((0, H))

        if not pixels:
            if isunit != 6:
                u = sunit[isunit]
                if u == "u":
                    u = "µ"
                ax.set_xlabel(f"x [{u}{unit}]")
                ax.set_ylabel(f"y [{u}{unit}]")
            else:
                ax.set_xlabel(f"x [{unit}]")
                ax.set_ylabel(f"y [{unit}]")
        if title is not None:
            ax.set_title(title)
        return r

    def real2px(self, x, y):
        """
        Transform a real (x,y) value in pixels
        Units should be the same as the one plotted by pySPM.SPM_image.show
        """
        return self.real2pixels(x, y)

    def real2pixels(self, x, y, float=False):
        """
        Transform a real (x,y) value in pixels
        Units should be the same as the one plotted by pySPM.SPM_image.show
        """
        W = self.size["real"]["x"]
        fact = int(np.floor(np.log(W) / np.log(10) / 3)) * 3
        if not float:
            px = np.digitize(
                x,
                np.linspace(
                    0, self.size["real"]["x"] / (10**fact), self.pixels.shape[1]
                ),
                right=True,
            )
            py = np.digitize(
                y,
                np.linspace(
                    0, self.size["real"]["y"] / (10**fact), self.pixels.shape[0]
                ),
                right=False,
            )
        else:
            px = x * (self.pixels.shape[1] - 1) / (self.size["real"]["x"] / (10**fact))
            py = y * (self.pixels.shape[0] - 1) / (self.size["real"]["y"] / (10**fact))
        return px, py

    def px2real(self, x, y):
        """
        Transform  a (x,y) value from pixels to real
        Units are the same as the one plotted by pySPM.SPM_image.show
        """
        W = self.size["real"]["x"]
        fact = int(np.floor(np.log(W) / np.log(10) / 3)) * 3
        rx = x * self.size["real"]["x"] / (10**fact) / self.pixels.shape[1]
        ry = (
            (self.pixels.shape[0] - y)
            * self.size["real"]["y"]
            / (10**fact)
            / self.pixels.shape[0]
        )
        return rx, ry

    def circular_profile(
        self,
        x0,
        y0,
        Ra=1,
        Rn=0,
        width=1,
        N=20,
        A=0,
        B=360,
        cmap="jet",
        axImg=None,
        axPolar=None,
        axProfile=None,
        plotProfileEvery=1,
        xtransf=lambda x: x * 1e9,
        ytransf=lambda x: x * 1e9,
        ToFcorr=False,
        fit=lambda x, *p: p[3] + p[2] * CDF(x, *p[:2]),
        p0=None,
        errors=False,
        bounds=(-np.inf, np.inf),
        fakefit=False,
        **kargs,
    ):
        """
        Create radial profiles from point x0,y0 with length Ra (outer radius) and Rn (negative Radius).
        Start from angle A° to angle B° with N profiles.
        If you want to apply the ToF-correction, please set ToFcorr to the number of scans used to record the ToF-SIMS image.
        Return the fitting uncertainty on sigma if errors is set to True
        The fitting function can be adjusted by fit and the default parameters by p0 which is an array of function where the first parameter passed will be the x-values and the second the y-values.
        """
        from matplotlib import cm, colors

        # Create a colormap for each profile
        CM = plt.get_cmap(cmap)
        cNorm = colors.Normalize(vmin=0, vmax=N)
        scalarMap = cm.ScalarMappable(norm=cNorm, cmap=CM)
        res = []
        cov = []
        angles = []
        assert A < B
        for i, angle in enumerate(np.linspace(A, B, N)):
            a = np.radians(angle)
            angles.append(a)
            l, p = self.get_profile(
                x0 - Rn * np.cos(a),
                y0 + Rn * np.sin(a),
                x0 + Ra * np.cos(a),
                y0 - Ra * np.sin(a),
                ax=axImg,
                width=width,
                color=scalarMap.to_rgba(i),
                **kargs,
            )
            profile = p if width == 0 else np.mean(p, axis=1)
            if ToFcorr:
                profile = -np.log(1.001 - profile / ToFcorr)
            if p0 is None:
                AC = np.mean(profile[: len(l) // 2])
                AE = np.mean(profile[len(l) // 2 :])
                if AC < AE:
                    p0 = [
                        l[len(l) // 2],
                        5 * (l[1] - l[0]),
                        np.max(profile) - np.min(profile),
                        np.min(profile),
                    ]
                else:
                    p0 = [
                        l[len(l) // 2],
                        5 * (l[1] - l[0]),
                        -np.max(profile) + np.min(profile),
                        np.max(profile),
                    ]
            else:
                for j, p in enumerate(p0):
                    if callable(p):
                        p0[j] = p(l, profile)
            if kargs.get("debug", False):
                print("calculate fit parameters are", p0)
            if not fakefit:
                p0, pcov = scipy.optimize.curve_fit(fit, l, profile, p0)
            else:
                pcov = np.zeros((len(p0), len(p0)))
            res.append(p0)
            cov.append([np.sqrt(abs(pcov[i, i])) for i in range(len(p0))])
            if axProfile and i % plotProfileEvery == 0:
                axProfile.plot(
                    xtransf(l - p0[0]),
                    profile,
                    color=scalarMap.to_rgba(i),
                    linestyle=":",
                )
                axProfile.plot(
                    xtransf(l - p0[0]), fit(l, *p0), color=scalarMap.to_rgba(i)
                )
        # close loop
        if A % 360 == B % 360:
            angles.append(angles[0])
            res.append(res[0])
            cov.append(cov[0])

        # Plot polar
        angles = np.array(angles)
        res = np.array(res)
        cov = np.array(cov)
        fact = 2 * np.sqrt(2 * np.log(2))
        if axPolar:
            axPolar.plot(
                angles,
                ytransf(res[:, 1]),
                color=kargs.get("sig_color", "C0"),
                label="$\\sigma$",
            )
            axPolar.plot(
                angles,
                ytransf(fact * res[:, 1]),
                color=kargs.get("fwhm_color", "C1"),
                label="FWHM",
            )
            if errors:
                axPolar.fill_between(
                    angles,
                    ytransf(res[:, 1] - cov[:, 1]),
                    ytransf(res[:, 1] + cov[:, 1]),
                    color=kargs.get("sig_color", "C0"),
                    alpha=kargs.get("fillalpha", 0.5),
                )
                axPolar.fill_between(
                    angles,
                    fact * ytransf(res[:, 1] - cov[:, 1]),
                    ytransf(res[:, 1] + cov[:, 1]),
                    color=kargs.get("fwhm_color", "C1"),
                    alpha=kargs.get("fillalpha", 0.5),
                )

        return angles, res, cov

    def get_profile(
        self,
        x1,
        y1,
        x2,
        y2,
        width=0,
        ax=None,
        pixels=True,
        color="w",
        axPixels=None,
        **kargs,
    ):
        """
        retrieve the profile of the image between pixel x1,y1 and x2,y2

        Parameters
        ----------
        x1, y1, x2, y2 : ints
            coordinates for the profile
        ax : matplotlib axis
            defines the matplotlib axis on which the position of the profile should be drawn (in not None)
        width : int
            the width of the profile (for averaging/statistics) in pixels
        color : string
            color used to plot the profiles lines
        axPixels : bool
            If True the image plotted in the ax axis is displayed in pixels

        Returns
        -------
        x data : 1D numpy array
        profile : 1D numpy array
        """
        if kargs.get("debug", False):
            print("get_profile input coordinates:", x1, y1, x2, y2)
        if ax is not None and axPixels is None and hasattr(ax, "isPixel"):
            axPixels = ax.isPixel
        if axPixels is None:
            axPixels = pixels
        W = self.size["real"]["x"]
        fact = int(np.floor(np.log(W) / np.log(10) / 3)) * 3
        if not pixels:
            if kargs.get("debug", False):
                print(
                    "Image range (real scale):",
                    self.size["real"]["x"] / (10**fact),
                    self.size["real"]["y"] / (10**fact),
                )
            x1, y1 = self.real2pixels(x1, y1, float=True)
            x2, y2 = self.real2pixels(x2, y2, float=True)
            y1 = self.pixels.shape[0] - y1
            y2 = self.pixels.shape[0] - y2
            if kargs.get("debug", False):
                print("Pixel coordinates:", x1, y1, x2, y2)
            if not axPixels:
                xvalues, p = get_profile(
                    np.flipud(self.pixels),
                    x1,
                    y1,
                    x2,
                    y2,
                    ax=ax,
                    width=width,
                    color=color,
                    transx=lambda x: x
                    * (self.size["real"]["x"] / (10**fact))
                    / self.pixels.shape[1],
                    transy=lambda x: (self.pixels.shape[0] - x)
                    * (self.size["real"]["y"] / (10**fact))
                    / self.pixels.shape[0],
                    **kargs,
                )
            else:
                values, p = get_profile(
                    np.flipud(self.pixels),
                    x1,
                    y1,
                    x2,
                    y2,
                    ax=ax,
                    width=width,
                    color=color,
                    **kargs,
                )
        else:
            if axPixels:
                values, p = get_profile(
                    np.flipud(self.pixels),
                    x1,
                    y1,
                    x2,
                    y2,
                    ax=ax,
                    width=width,
                    color=color,
                    **kargs,
                )
            else:
                values, p = get_profile(
                    np.flipud(self.pixels),
                    x1,
                    y1,
                    x2,
                    y2,
                    ax=ax,
                    width=width,
                    color=color,
                    transx=lambda x: x
                    * (self.size["real"]["x"] / (10**fact))
                    / self.pixels.shape[1],
                    transy=lambda x: (self.pixels.shape[0] - x)
                    * (self.size["real"]["y"] / (10**fact))
                    / self.pixels.shape[0],
                    **kargs,
                )

        dx = (x2 - x1) * self.size["real"]["x"] / self.size["pixels"]["x"]
        dy = (y2 - y1) * self.size["real"]["y"] / self.size["pixels"]["y"]
        rd = np.sqrt(dx**2 + dy**2)
        xvalues = np.linspace(0, rd, len(p))
        return xvalues, p

    def plot_profile(
        self,
        x1,
        y1,
        x2,
        y2,
        width=0,
        ax=None,
        pixels=True,
        img=None,
        imgColor="w",
        ztransf=lambda x: x,
        zunit=None,
        **kargs,
    ):
        """
        Retrieve and plot a profile from an image

        Parameters
        ----------
        x1, y1, x2, y2 : int
            coordinate of the profile in real size or in pixels (if pixels is True)
        width : float
            the width of the profiles in pixels for better statistics
        ax : matplotlib axis
            The axis in which the profile will be plotted
        pixels : bool
            If True the coordinates are given in pixels and not in real units
        img : matplotlib axis
            The axis in which the profile position will be drawn
        imgColor : string
            The color used to display the profile positions
        ztransf : function
            function to transform the profile data. This can be used to scale the data.
            Most profiles are retrieved in 'm' and a 'nm' value can be used by using ztransf=lambda x: x*1e9
        zunit : string
            the zunit name used if ztransft is used
        color : string
            The color of the profile
        col : string
            can be used instead of color
        stdplot : bool
            If True display the ±nσ plots where n is given by the sig parameter
        sig : int
            The number of sigmas used in stdplot
        label : string
            The label used for plotting the profile (useful if you perform a ax.legend() afterwards)

        Returns
        -------
        dictionary : {'plot': matplotlib_plot_instance, 'l': profile_xaxis, 'z': profile_yaxis}

        Examples
        --------
        >>> topo = pySPM.SPM_image(...)
        >>> fig, ax = plt.subplots(1, 2, figsize=(10, 5))
        >>> topo.plot_profile(70, 100, 170, 200, ax=ax[1], img=ax[0], ztransf=lambda x:x*1e9, zunit='nm');
        >>> topo.show(ax=ax[0], pixels=True);
        """
        col = kargs.get("color", kargs.get("col", "C0"))
        W = self.size["real"]["x"]
        fact = int(np.floor(np.log(W) / np.log(10) / 3)) * 3
        if ax is None:
            ax = plt.gca()
        xvalues, p = self.get_profile(
            x1, y1, x2, y2, width=width, color=imgColor, ax=img, pixels=pixels, **kargs
        )
        d = np.sqrt((x2 - x1) ** 2 + (y2 - y1) ** 2)
        dx = x2 - x1
        dy = y2 - y1
        if pixels:
            rd = d
            u = ""
            unit = "px"
        else:
            unit = self.size["real"]["unit"]
            sunit = "afpnum kMGTPE"
            if len(unit) == 1:
                isunit = 6
            elif unit[0] in sunit:
                isunit = sunit.find(unit[0])
                unit = unit[1:]
            else:
                isunit = 6
            isunit += fact // 3
            u = sunit[isunit] if isunit != 6 else ""
            if u == "u":
                u = "µ"
            rd = np.sqrt(dx**2 + dy**2)
        xvalues = np.linspace(0, rd, len(p))
        lab = kargs.get("label", "")
        if width < 2:
            profile = ztransf(p)
        else:
            profile = ztransf(np.mean(p, axis=1))
            s = np.std(p)
            if kargs.get("stdplot", False):
                for ns in range(1, kargs.get("sig", 2) + 1):
                    ax.fill_between(
                        xvalues,
                        profile - ns * s,
                        profile + ns * s,
                        color=col,
                        alpha=0.2,
                        label=[
                            lab
                            + r" ($\sigma,\ldots {}\sigma$)".format(
                                kargs.get("sig", 2)
                            ),
                            None,
                        ][ns > 1],
                    )

        Plot = ax.plot(
            xvalues,
            profile,
            color=col,
            linewidth=kargs.get("linewidth", 1),
            linestyle=kargs.get("linestyle", "-"),
            label=lab + [" (mean)", ""][width < 2],
        )
        if kargs.get("min", False):
            minStyle = kargs.get("minStyle", kargs.get("minmaxStyle", "--"))
            minColor = kargs.get("minColor", kargs.get("minmaxColor", col))
            minMarker = kargs.get("minMarker", kargs.get("minmaxMarker", ""))
            ax.plot(
                xvalues,
                np.min(p, axis=1),
                color=minColor,
                linewidth=kargs.get("linewidth", 1),
                linestyle=minStyle,
                marker=minMarker,
                label=lab + " (min)",
            )
        if kargs.get("max", False):
            maxStyle = kargs.get("maxStyle", kargs.get("minmaxStyle", "--"))
            maxColor = kargs.get("maxColor", kargs.get("minmaxColor", col))
            maxMarker = kargs.get("maxMarker", kargs.get("minmaxMarker", ""))
            ax.plot(
                xvalues,
                np.max(p, axis=1),
                color=maxColor,
                linestyle=maxStyle,
                linewidth=kargs.get("linewidth", 1),
                marker=maxMarker,
                label=lab + " (max)",
            )

        ax.set_xlabel(f"Distance [{u}{unit}]")
        if zunit is not None:
            ax.set_ylabel(f"{self.channel} [{zunit}]")
        else:
            ax.set_ylabel(f"{self.channel} [{self.zscale}]")

        return {"plot": Plot, "l": xvalues, "z": profile}

    def get_bin_threshold(
        self, percent, high=True, adaptive=False, binary=True, img=False
    ):
        """
        Threshold the image into binary values

        Parameters
        ----------
        percent : float
            The percentage where the thresholding is made
        high : bool
            If high a value of 1 is returned for values > percent
        adaptive : bool
            If True, performs an adaptive thresholding (see skimage.filters.threshold_adaptive)
        binary : bool
            If True return bool data (True/False) otherwise  numeric (0/1)
        img : bool
            If True return a SPM_image otherwise a numpy array
        """
        if adaptive:
            if binary:
                return self.pixels > threshold_local(self.pixels, percent)
            return threshold_local(self.pixels, percent)
        mi = np.min(self.pixels)
        norm = (self.pixels - mi) / (np.max(self.pixels) - mi)
        r = norm > percent if high else norm < percent
        if not img:
            if binary:
                return r
            return np.ones(self.pixels.shape) * r
        else:
            I = copy.deepcopy(self)
            I.channel = "Threshold from " + I.channel
            if binary:
                I.pixels = r
            else:
                I.pixels = np.ones(self.pixels.shape) * r
            return I

    def spline_offset(self, X, Y, Z=None, inline=True, ax=None, output="img", **kargs):
        """
        subtract a spline interpolated by points corrdinates.
        if Z is None, the image values will be used (default)
        """
        if ax is not None:
            if "num" in kargs and kargs["num"]:
                text_color = "k"
                if "text_color" in kargs:
                    text_color = kargs["text_color"]
                    del kargs["text_color"]
                for i in range(len(X)):
                    l = self.pixels.shape[1] - X[i] < 20
                    ax.annotate(
                        str(i),
                        (X[i], Y[i]),
                        ([5, -5][l], 0),
                        textcoords="offset pixels",
                        va="center",
                        ha=["left", "right"][l],
                        color=text_color,
                    )
                del kargs["num"]
            ax.plot(X, Y, "o", **kargs)
        import scipy.interpolate

        T = np.flipud(self.pixels) - np.min(self.pixels)
        if Z is None:
            Z = [T[Y[i], X[i]] for i in range(len(X))]
        x = np.arange(self.pixels.shape[1])
        y = np.arange(self.pixels.shape[0])
        xx, yy = np.meshgrid(x, y)
        I = scipy.interpolate.SmoothBivariateSpline(X, Y, Z)
        z = I.ev(xx, yy)
        if inline:
            self.pixels -= z
            return z
        else:
            if output == "img":
                New = copy.deepcopy(self)
                New.pixels -= z
                return New
            elif output == "spline":
                return z
            else:
                raise ValueError(
                    "The output parameter should be either 'img' or 'spline'"
                )

    def get_shadow_mask(self, angle, BIN=None, prog=False):
        """
        If an image is recorded with a beam incident with a certain angle, the topography will shadow the data.
        This function generates the shadow mask for a given topography and a given incident angle.

        Parameters
        ----------
        angle : float
            The incidence angle in degrees
        BIN : numpy array
            Data. If given will move the recorded pixels at the correct x,y positions
        prog : bool
            display a progressbar ?

        Note
        ----
        This function is old, might not be optimized or working properly
        """
        if BIN is not None:
            BIN = BIN * 1.0
        slope = np.tan(np.radians(angle))
        neg = False
        if slope < 0:
            neg = True
            slope = -slope
            topo = np.fliplr(self.pixels)
            if BIN is not None:
                BIN = np.fliplr(BIN)
        else:
            topo = self.pixels
        x = np.linspace(0, self.size["real"]["x"], self.pixels.shape[1])
        if self.size["real"]["unit"] == "um":
            x *= 1e-6
        elif self.size["real"]["unit"] == "nm":
            x *= 1e-9
        mask = np.zeros(self.pixels.shape)
        AFM_bin_shadow = np.zeros(self.pixels.shape)
        Y = range(self.pixels.shape[0])
        if prog:
            Y = PB(Y)
        for yi in Y:
            for xi in range(self.pixels.shape[1]):
                cut = self.pixels.shape[1] - 2
                y_ray = slope * (x - x[xi]) + topo[yi, xi]
                while cut > xi and y_ray[cut] > topo[yi, cut]:
                    cut -= 1
                if xi == cut:
                    if BIN is not None:
                        AFM_bin_shadow[yi, xi] = BIN[yi, xi]
                    continue
                # Cut has been found
                if BIN is not None:
                    x1 = x[cut]
                    x2 = x[cut + 1]
                    y1 = topo[yi, cut]
                    y2 = topo[yi, cut + 1]
                    x0 = x[xi]
                    y0 = topo[yi, xi]
                    if y2 == y1:
                        x_cut = (y1 + slope * x0 - y0) / slope
                    else:
                        numerator = x1 / (x2 - x1) + (y0 - slope * x0 - y1) / (y2 - y1)
                        denominator = 1 / (x2 - x1) - slope / (y2 - y1)
                        x_cut = numerator / denominator
                        slope * (x_cut - x0) + y0
                    if x_cut >= x1 and x_cut <= x2:
                        y1 = BIN[yi, cut]
                        y2 = BIN[yi, cut + 1]
                        yint = (((y2 - y1) / (x2 - x1)) * (x_cut - x1)) + y1
                    else:
                        yint = BIN[yi, xi]
                    AFM_bin_shadow[yi, xi] = yint
                mask[yi, xi] = 1
        if neg:
            mask = np.fliplr(mask)
            AFM_bin_shadow = np.fliplr(AFM_bin_shadow)
        if BIN is not None:
            return (mask, AFM_bin_shadow)
        return mask

    def adjust_position(self, fixed):
        """
        Shift the current pixels to match a fixed image.
        The shift is determined by position where the cross-correlation is maximized.
        """
        adj = copy.deepcopy(self)
        cor = np.fft.fft2(fixed.pixels)
        cor = np.abs(np.fft.ifft2(np.conj(cor) * np.fft.fft2(self.pixels)))
        cor = cor / fixed.pixels.size
        ypeak, xpeak = np.unravel_index(cor.argmax(), cor.shape)
        shift = [-(ypeak - 1), -(xpeak - 1)]
        adj.pixels = np.roll(self.pixels, shift[0], axis=0)
        adj.pixels = np.roll(adj.pixels, shift[1], axis=1)
        return adj

    def align(self, tform, cut=True):
        """
        Apply an Affine transform on the data

        Parameters
        ----------
        tform : skimage.transform
            the affine transform to perform
        cut : bool
            If True cut the data
        """
        New = copy.deepcopy(self)
        New.pixels = tf.warp(self.pixels, tform, preserve_range=True)
        if not cut:
            return New
        cut = [0, 0, *list(self.pixels.shape)]
        if tform.translation[0] >= 0:
            cut[2] -= tform.translation[0]
        elif tform.translation[0] < 0:
            cut[0] -= tform.translation[0]
        if tform.translation[1] >= 0:
            cut[1] += tform.translation[1]
        elif tform.translation[1] < 0:
            cut[3] += tform.translation[1]
        cut = [int(x) for x in cut]
        New.cut(cut, inplace=True)
        return New, cut

    def get_fft(self):
        """
        return the FFT2 transform opf the image
        """
        return np.fft.fftshift(np.fft.fft2(self.pixels))

    def corr_fit2d(self, nx=2, ny=1, poly=False, inline=True, mask=None):
        """
        Subtract a fitted 2D-polynom of nx and ny order from the data

        Parameters
        ----------
        nx : int
            the polynom order for the x-axis
        ny : int
            the polynom order for the y-axis
        poly : bool
            if True the polynom is returned as output
        inline : bool
            create a new object?
        mask : 2D numpy array
            mask where the fitting should be performed
        """
        r, z = fit2d(self.pixels, nx, ny, mask=mask)
        if inline:
            self.pixels -= z
        else:
            N = copy.deepcopy(self)
            N.pixels -= z
            if poly:
                return N, z
            return N
        if poly:
            return z
        return self

    def zero_min(self, inline=True):
        """
        Shift the values so that the minimum becomes zero.
        """
        if inline:
            self.pixels -= np.min(self.pixels)
            return self
        else:
            N = copy.deepcopy(self)
            N.pixels -= np.min(N.pixels)
            return N

    def get_radius_mask_from_center(self, radius):
        mask = np.zeros(self.pixels.shape, dtype=np.uint8)
        rr, cc = disk(
            center=(self.pixels.shape[0] // 2, self.pixels.shape[1] // 2), radius=radius
        )
        mask[rr, cc] = 1
        return mask

    def filter_lowpass(self, fft_radius, inline=True):
        """
        Execute a lowpass filter on the data
        """
        F = self.get_fft()
        mask = self.get_radius_mask_from_center(radius=fft_radius)
        if inline:
            self.pixels = np.real(np.fft.ifft2(np.fft.fftshift(F * mask)))
        else:
            C = copy.deepcopy(self)
            C.pixels = np.real(np.fft.ifft2(np.fft.fftshift(F * mask)))
            return C

    def filter_gaussian(self, sigma, inline=True, **kwargs):
        """
        Execute a gaussian filter on the data
        """
        if inline:
            self.pixels = scipy.ndimage.gaussian_filter(
                self.pixels, sigma=sigma, **kwargs
            )
        else:
            C = copy.deepcopy(self)
            C.pixels = scipy.ndimage.gaussian_filter(C.pixels, sigma=sigma, **kwargs)
            return C

    def filter_percentile(self, percentile, size, inline=True, **kwargs):
        """
        Execute a percentile filter on the data
        """
        if inline:
            self.pixels = scipy.ndimage.percentile_filter(
                self.pixels, percentile=percentile, size=size, **kwargs
            )
        else:
            C = copy.deepcopy(self)
            C.pixels = scipy.ndimage.percentile_filter(
                C.pixels, percentile=percentile, size=size, **kwargs
            )
            return C

    def _resize_infos(self):
        """
        Internal to recalculate the real size when the image is cropped or cut
        """
        self.size["real"]["x"] *= self.pixels.shape[1] / self.size["pixels"]["x"]
        self.size["real"]["y"] *= self.pixels.shape[0] / self.size["pixels"]["y"]
        self.size["pixels"]["x"] = int(self.pixels.shape[1])
        self.size["pixels"]["y"] = int(self.pixels.shape[0])
        if "recorded" in self.size:
            self.size["recorded"]["real"]["x"] *= (
                self.pixels.shape[1] / self.size["pixels"]["x"]
            )
            self.size["recorded"]["real"]["y"] *= (
                self.pixels.shape[0] / self.size["pixels"]["y"]
            )
            self.size["recorded"]["pixels"]["x"] = int(self.pixels.shape[1])
            self.size["recorded"]["pixels"]["y"] = int(self.pixels.shape[0])

    def filter_scars_removal(self, thresh=0.5, inline=True):
        """
        Filter function to remove scars from images.
        """
        C = copy.deepcopy(self) if not inline else self
        for y in range(1, self.pixels.shape[0] - 1):
            b = self.pixels[y - 1, :]
            c = self.pixels[y, :]
            a = self.pixels[y + 1, :]
            mask = np.abs(b - a) < thresh * (np.abs(c - a))
            C.pixels[y, mask] = b[mask]
        if not inline:
            return C
        return self

    def cut(self, c, inline=False, pixels=True, **kargs):
        """
        Clip/Crop the image

        Parameters
        ----------
        c : list [llx,lly,urx,ury]
            list of the lowe-left (ll) and upper-right (ur) coordinates
        inline: bool
            perform the transformation inline or produce a new SPM_image?
        pixels : bool
            Are the coordinates given in pixels?

        Returns
        -------
        self if inplace, clipped SPM_image otherwises2 = pySPM.Nanoscan("%s/CyI5b_PCB_ns.xml"%(Path))
        """
        if "inplace" in kargs:
            inline = kargs["inplace"]
        if kargs.get("debug", False):
            print("cut) Input coordinates:", c)
        if not pixels:
            c = [z for s in zip(*self.real2pixels(c[0::2], c[1::2])) for z in s]
            if kargs.get("debug", False):
                print("cut) pixel coordinates:", c)
        if not inline:
            new = copy.deepcopy(self)
            new.pixels = cut(self.pixels, c, **kargs)
            new._resize_infos()
            return new
        else:
            self.pixels = cut(self.pixels, c, **kargs)
            self._resize_infos()
            return self

    def zoom(self, zoom_factor, inplace=False, order=3):
        """
        Resize the image to a new pixel size (but keep the real size) by pixel interpolation.

        Parameters
        ----------
        zoom_factor : float
            > 1: up sampling
            < 1: down sampling
        order : int
            The spline interpolation order to use. (default: 3). Use 0 for binary or very sharp images.
        inplace : bool
            create a new image?
        """
        from scipy.ndimage.interpolation import zoom

        if not inplace:
            new = copy.deepcopy(self)
            new.pixels = zoom(new.pixels, zoom_factor, order=order)
            new.size["pixels"]["x"] = new.pixels.shape[1]
            new.size["pixels"]["y"] = new.pixels.shape[0]
            return new
        else:
            self.pixels = zoom(self.pixels, zoom_factor, order=order)
            self.size["pixels"]["x"] = self.pixels.shape[1]
            self.size["pixels"]["y"] = self.pixels.shape[0]
            return self


# Note: The following functions are not part of the SPM_image class.
# All following functions are performed on numpy arrays


def cut(img, c, **kargs):
    """
    Clip / Crop a numpy array

    Parameters
    ----------
    img : 2D numpy array
        The input image array
    c : list [llx, lly, urx, ury]
        the lower-left (ll) and upper-right (ur) coordinates used for the cropping
    """
    from .utils.geometry import Bbox

    if kargs.get("debug", False):
        print("cut in x", c[0], "->", c[2], " - in y", c[1], "->", c[3])
    if isinstance(c, Bbox):
        c = [c.left, c.bottom, c.right, c.top]
    if c[3] < c[1]:
        c = [c[0], c[3], c[2], c[1]]
    if c[2] < c[0]:
        c = [c[2], c[1], c[0], c[3]]
    if c[2] - c[0] == img.shape[1] and c[3] - c[1] == img.shape[0]:
        raise Exception("Reshaping the same array again?")
    return img[c[1] : c[3], c[0] : c[2]]


def normalize(data, sig=None, vmin=None, vmax=None):
    """
    Normalize the input data. Minimum_value -> 0 and maximum_value -> 1

    Parameters
    ----------
    data : numpy array
        input data
    sig : float or None
        if not None:
            mean(data)-sig*standard_deviation(data) -> 0
            mean(data)+sig*standard_deviation(data) -> 1
    vmin : float or None
        if not None, define the lower bound i.e.  vmin -> 0
    vmax : float or None
        if not None, defines the upper bound i.e. vmax -> 0

    Note
    ----
    All values below the lower bound will be = 0
    and all values above the upper bound will be = 1


    """
    if sig is None:
        mi = np.min(data)
        ma = np.max(data)
    else:
        s = sig * np.std(data)
        mi = np.mean(data) - s
        ma = np.mean(data) + s
    if vmin is not None:
        mi = vmin
    if vmax is not None:
        ma = vmax
    N = (data - mi) / (ma - mi)
    N[N < 0] = 0
    N[N > 1] = 1
    return N


def imshow_sig(img, sig=1, ax=None, **kargs):
    """
    Shortcut to plot a numpy array around it's mean with bounds ±sig sigmas

    Parameters
    ----------
    img : 2D numpy array
        input image to display
    sig : float
        The number of standard-deviation to plot
    ax : matplotlib axis
        matplotlib axis to use. If None, the current axis (plt.gca() will be used).
    **kargs : additional parameters
        will be passed to the imshow function of matplotls2 = pySPM.Nanoscan("%s/CyI5b_PCB_ns.xml"%(Path))ib
    """
    if ax is None:
        fig, ax = plt.subplots(1, 1)
    std = np.std(img)
    avg = np.mean(img)
    vmin = avg - sig * std
    vmax = avg + sig * std
    ax.imshow(img, vmin=vmin, vmax=vmax, **kargs)


def adjust_position(fixed, to_adjust, shift=False):
    """Shift the current pixels to match a fixed image by rolling the data"""
    adj = copy.deepcopy(to_adjust)
    cor = np.fft.fft2(fixed)
    cor = np.abs(np.fft.ifft2(np.conj(cor) * np.fft.fft2(to_adjust)))
    cor = cor / to_adjust.size
    ypeak, xpeak = np.unravel_index(cor.argmax(), cor.shape)
    shift = [-(ypeak - 1), -(xpeak - 1)]
    adj = np.roll(to_adjust, shift[0], axis=0)
    adj = np.roll(adj, shift[1], axis=1)
    if shift:
        return adj, shift
    return adj


def tukeyfy(A, alpha, type="default"):
    """
    Apply a Tukey window on the current image

    Parameters
    ----------
    A : 2D numpy array
        input array
    alpha : float
        Size of the Tukey windows in percent of the image (≥0 and ≤1)
    type : string
        if not "default" perform a mean centering (data will blend down to its mean instead of 0)
    """
    tuky = tukeywin(A.shape[0], alpha)
    tukx = tukeywin(A.shape[1], alpha)
    tuk = np.multiply(tukx[:, None].T, tuky[:, None])
    if type == "default":
        return A * tuk
    avg = np.mean(A)
    return avg + (A - avg) * tuk


def tukeywin(window_length, alpha=0.5):
    """The Tukey window, also known as the tapered cosine window, can be regarded as a cosine lobe of width \alpha * N / 2
    that is convolved with a rectangle window of width (1 - \alpha / 2). At \alpha = 1 it becomes rectangular, and
    at \alpha = 0 it becomes a Hann window.

    We use the same reference as MATLAB to provide the same results in case users compare a MATLAB output to this function
    output

    Reference
    ---------
    http://www.mathworks.com/access/helpdesk/help/toolbox/signal/tukeywin.html

    """
    # Special cases
    if alpha <= 0:
        return np.ones(window_length)  # rectangular window
    elif alpha >= 1:
        return np.hanning(window_length)

    # Normal case
    x = np.linspace(0, 1, window_length)
    w = np.ones(x.shape)

    # first condition 0 <= x < alpha/2
    first_condition = x < alpha / 2
    w[first_condition] = 0.5 * (
        1 + np.cos(2 * np.pi / alpha * (x[first_condition] - alpha / 2))
    )

    # second condition already taken care of

    # third condition 1 - alpha / 2 <= x <= 1
    third_condition = x >= (1 - alpha / 2)
    w[third_condition] = 0.5 * (
        1 + np.cos(2 * np.pi / alpha * (x[third_condition] - 1 + alpha / 2))
    )
    return w


def overlay(ax, mask, color, **kargs):
    """
    Plot an overlay on an existing axis

    Parameters
    ----------
    ax : matplotlib axis
        input axis
    mask : 2D numpy array
        Binary array where a mask should be plotted
    color : string
        The color of the mask to  plot
    **kargs: additional parameters
        passed to the imshow function of matploltib
    """
    m = np.ma.masked_array(mask, ~mask)
    col = np.array(mpl.colors.colorConverter.to_rgba(color))
    I = col[:, None, None].T * m[:, :, None]
    ax.imshow(I, **kargs)


def normP(x, p, trunk=True):
    """
    Normalize the input data accroding to its percentile value.

    Parameters
    ----------
    x : 2D numpy array
        input data
    p : float
        percentile to normalize the data.
        lower bound = p percentile
        upper bound = (100-p) percentile
    trunk : bool
        If True the data are truncated between 0 and 1
    """
    thresh_high = np.percentile(x, 100 - p)
    thresh_low = np.percentile(x, p)
    if thresh_low == thresh_high:
        thresh_high = np.max(x)
        thresh_low = np.min(x)
    if thresh_low == thresh_high:
        thresh_high = thresh_low + 1
    r = (x - thresh_low) / (thresh_high - thresh_low)
    if trunk:
        r[r < 0] = 0
        r[r > 1] = 1
    return r


def beam_profile(
    target,
    source,
    mu=1e-6,
    tukey=0,
    meanCorr=False,
    source_tukey=None,
    real=np.abs,
    **kargs,
):
    """
    Calculate the PSF by deconvolution of the target
    with the source using a Tikhonov regularization of factor mu.
    """
    if source_tukey is None:
        source_tukey = tukey
    if kargs.get("source_centering", False):
        source = 2 * source - 1
    if meanCorr:
        target = target - np.mean(target)
    if tukey > 0:
        target = tukeyfy(target, tukey)
    if source_tukey > 0:
        source = tukeyfy(source, tukey)
    tf = np.fft.fft2(source)
    tf /= np.size(tf)
    recon_tf = np.conj(tf) / (np.abs(tf) ** 2 + mu)
    return np.fft.fftshift(
        real(np.fft.ifft2(np.fft.fft2(target) * recon_tf))
    ) / np.size(target)


def beam_profile1d(target, source, mu=1e-6, real=np.abs):
    source = source
    tf = np.fft.fft(source)
    tf /= np.size(tf)
    recon_tf = np.conj(tf) / (np.abs(tf) ** 2 + mu)
    F = np.fft.fft(target) * recon_tf
    return np.fft.fftshift(real(np.fft.ifft(F))), F


def zoom_center(img, sx, sy=None):
    """
    Zoom by taking the sx × sy central pixels

    Parameters
    ----------
    img : 2D numpy array
        The input data
    sx : int
        The number of pixels along the x-axis to take

    sy : int or None
        The number of pixels alongs the y-axis to take.
        If None take the same value as for sx
    """
    if sy is None:
        sy = sx
    assert isinstance(sx, int)
    assert isinstance(sy, int)
    return img[
        img.shape[0] // 2 - sy // 2 : img.shape[0] // 2 + sy // 2,
        img.shape[1] // 2 - sx // 2 : img.shape[1] // 2 + sx // 2,
    ]


def px2real(x, y, size, ext):
    rx = ext[0] + (x / size[1]) * (ext[1] - ext[0])
    ry = ext[2] + (y / size[0]) * (ext[3] - ext[2])
    return rx, ry


def real2px(x, y, size, ext):
    px = size[1] * (x - ext[0]) / (ext[1] - ext[0])
    py = size[0] * (y - ext[2]) / (ext[3] - ext[2])
    return px, py


def fit2d(Z0, dx=2, dy=1, mask=None):
    """
    Fit the input data with a 2D polynom of order dx × dy

    Parameters
    ----------
    Z0 : 2D numpy array
        input data
    dx : int
        order of the polynom for the x-axis
    dy : int
        order of the polynom for the y-xis
    mask : 2D numpy array
        Give a mask where True values only will be used to perform the fitting

    Returns
    -------
    numpy array
        fitting parameters
    2D numpy array
        result of the polynom
    """
    x = np.arange(Z0.shape[1], dtype=float)
    y = np.arange(Z0.shape[0], dtype=float)
    X0, Y0 = np.meshgrid(x, y)
    if mask is not None:
        X = X0[mask]
        Y = Y0[mask]
        Z = Z0[mask]
    else:
        X = X0
        Y = Y0
        Z = Z0
    x2 = X.ravel()
    y2 = Y.ravel()
    A = np.vstack([x2**i for i in range(dx + 1)])
    A = np.vstack([A] + [y2**i for i in range(1, dy + 1)])
    res = scipy.optimize.lsq_linear(A.T, Z.ravel())
    r = res["x"]
    Z2 = r[0] * np.ones(Z0.shape)
    for i in range(1, dx + 1):
        Z2 += r[i] * (X0**i)
    for i in range(1, dy + 1):
        Z2 += r[dx + i] * (Y0**i)
    return r, Z2


def warp_and_cut(img, tform, cut=True):
    """
    Perform an Affine transform on the input data and cut them if cut=True

    Parameters
    ----------
    img : 2D numpy array
        input data
    tform : skimage.transform
        An Affine fransform to perform on the data
    cut : bool
        Should the data be cutted?
    """
    New = tf.warp(img, tform, preserve_range=True)
    Cut = [0, 0, *list(img.shape)]
    if tform.translation[0] >= 0:
        Cut[2] -= tform.translation[0]
    elif tform.translation[0] < 0:
        Cut[0] -= tform.translation[0]
    if tform.translation[1] >= 0:
        Cut[1] += tform.translation[1]
    elif tform.translation[1] < 0:
        Cut[3] += tform.translation[1]
    Cut = [int(x) for x in Cut]
    if cut:
        New = cut(New, Cut)
    return New, Cut


def get_profile(
    I,
    x1,
    y1,
    x2,
    y2,
    width=0,
    ax=None,
    color="w",
    alpha=0,
    N=None,
    transx=lambda x: x,
    transy=lambda x: x,
    interp_order=1,
    **kargs,
):
    """
    Get a profile from an input matrix.
    Low-level function. Doc will come laters2 = pySPM.Nanoscan("%s/CyI5b_PCB_ns.xml"%(Path))
    """
    d = np.sqrt((x2 - x1) ** 2 + (y2 - y1) ** 2)
    if N is None:
        N = int(d) + 1
    P = []
    dx = -width / 2 * (y2 - y1) / d
    dy = width / 2 * (x2 - x1) / d
    for w in np.linspace(-width / 2, width / 2, max(1, width)):
        dx = -w * (y2 - y1) / d
        dy = w * (x2 - x1) / d
        x = np.linspace(x1 + dx, x2 + dx, N)
        y = np.linspace(y1 + dy, y2 + dy, N)
        M = scipy.ndimage.interpolation.map_coordinates(
            I, np.vstack((y, x)), order=interp_order
        )
        P.append(M)
    if kargs.get("debug", False):
        print("get_profile input coordinates:", x1, y1, x2, y2)
    if ax is not None:
        x1 = transx(x1)
        x2 = transx(x2)
        y1 = transy(y1)
        y2 = transy(y2)
        if kargs.get("debug", False):
            print("Drawing coordinates:", x1, y1, x2, y2)
        dx = -width / 2 * (y2 - y1) / d
        dy = width / 2 * (x2 - x1) / d
        if type(color) in [tuple, list]:
            ax.plot(
                [x1, x2],
                [y1, y2],
                color=color,
                alpha=kargs.get("linealpha", 1),
                linestyle=kargs.get("linestyle", "-"),
            )
            ax.plot(
                [x1 - dx, x1 + dx],
                [y1 - dy, y1 + dy],
                color=color,
                alpha=kargs.get("linealpha", 1),
            )
            ax.plot(
                [x2 - dx, x2 + dx],
                [y2 - dy, y2 + dy],
                color=color,
                alpha=kargs.get("linealpha", 1),
            )
        else:
            ax.plot(
                [x1, x2],
                [y1, y2],
                color,
                alpha=kargs.get("linealpha", 1),
                lw=kargs.get("lw", 1),
                linestyle=kargs.get("linestyle", "-"),
            )
            ax.plot(
                [x1 - dx, x1 + dx],
                [y1 - dy, y1 + dy],
                color,
                alpha=kargs.get("linealpha", 1),
            )
            ax.plot(
                [x2 - dx, x2 + dx],
                [y2 - dy, y2 + dy],
                color,
                alpha=kargs.get("linealpha", 1),
            )
        if alpha > 0:
            import matplotlib.patches

            ax.add_patch(
                matplotlib.patches.Rectangle(
                    (x1 + dx, y1 + dy),
                    2 * np.sqrt(dx**2 + dy**2),
                    np.sqrt((x2 - x1) ** 2 + (y2 - y1) ** 2),
                    -np.degrees(np.arctan2(x2 - x1, y2 - y1)),
                    color=color,
                    alpha=alpha,
                )
            )
    if len(P) == 1:
        return np.linspace(0, d, N), P[0]
    return np.linspace(0, d, N), np.vstack(P).T


def dist_v2(img, dx=1, dy=1):
    """
    Return a 2D array with the distance in pixel with the clothest corner of the array.
    """
    x2 = np.arange(img.shape[1])
    x2 = (np.minimum(x2, img.shape[1] - x2) * dx) ** 2
    y2 = np.arange(img.shape[0])
    y2 = (np.minimum(y2, img.shape[0] - y2) * dy) ** 2
    X, Y = np.meshgrid(x2, y2)
    return np.sqrt(X + Y)


def generate_k_matrices(A, dx, dy):
    """
    GENERATE_K_MATRICES k-Matrix generation (helper function).
    generates k-matrices for the 2D-channel CHANNEL.

    K is a matrix of the same size as the pixel matrix A, containing the real-life frequency distance of each
    pixel position to the nearest corner of an matrix that is one pixel
    wider/higher.
    KX is of the same size as K and contains the real-life  difference in x-direction of each pixel position to the nearest corner
    of a matrix that is one pixel wider/higher.
    Similarly, KY is of the  same size as K, containing the real-life difference in y-direction of  each pixel position to the nearest corner of an matrix that is one
    pixel wider/higher.
    """
    ny, nx = A.shape
    dkx = 2 * np.pi / (nx * dx)
    dky = 2 * np.pi / (ny * dy)

    ky = np.arange(0, ny)
    ky = (np.mod(ky + ny / 2, ny) - ny / 2) * dky

    kx = np.arange(0, nx)
    kx = (np.mod(kx + nx / 2, nx) - nx / 2) * dkx

    kx, ky = np.meshgrid(kx, ky)
    k = dist_v2(A, dkx, dky)
    k[
        0, 0
    ] = 1.0  # Prevent division by zero error and illegal operand errors. This may be improved...
    return k, kx, ky


def mfm_tf(
    nx,
    dx,
    ny,
    dy,
    tf_in,
    derivative=0,
    transform=0,
    z=0,
    A=0,
    theta=None,
    phi=None,
    d=None,
    delta_w=None,
):
    """
    Draft for the MFM tf function
    """
    k, kx, ky = generate_k_matrices(tf_in, dx, dy)
    # Distance loss
    tf_out = np.exp(-z * k)
    if d is not None:
        tf_out = tf_out / 2.0
        if not np.isinf(d):
            if d == 0:
                tf_out *= k
            else:
                tf_out *= 1 - np.exp(-d * k)
    if A == 0:
        if transform != 0:
            assert theta is not None
            assert phi is not None
            tf_out *= (
                (
                    np.cos(theta)
                    + 1j
                    * (
                        np.cos(phi) * np.sin(-theta) * kx
                        + np.sin(phi) * np.sin(-theta) * ky
                    )
                )
                / k
            ) ** transform
        if derivative == 1:
            tf_out *= k
    else:
        pass  # TODO
    return tf_out * tf_in


def mfm_inv_calc_flat(
    img,
    z,
    tf_in,
    thickness=None,
    delta_w=None,
    amplitude=0,
    derivative=0,
    transform=0,
    mu=1e-8,
):
    """
    MFM inv calc function
    """
    theta = np.radians(12)
    phi = np.radians(-90)
    ny, nx = img.shape
    tf = mfm_tf(
        nx,
        1,
        ny,
        1,
        tf_in,
        derivative,
        transform,
        z,
        amplitude,
        theta,
        phi,
        thickness,
        delta_w,
    )
    tf[0, 0] = np.real(np.mean(tf))
    recon_tf = np.conj(tf) / (mu + np.abs(tf) ** 2)
    work_img = np.abs(np.fft.ifft2(np.fft.fft2(img) * recon_tf))
    return work_img


def get_tik_tf(Img, mu, tukey=0, source_tukey=0, debug=False, d=200, real=np.real):
    import scipy

    def fit(x, a, A, bg, x0):
        return bg + (A - bg) * np.exp(-abs(x - x0) / a)

    x = np.arange(Img.shape[1])
    y = np.arange(Img.shape[0])
    X, Y = np.meshgrid(x, y)
    x0 = Img.shape[1] / 2
    y0 = Img.shape[0] / 2
    R = np.sqrt((X - x0) ** 2 + (Y - y0) ** 2)

    Z = beam_profile(Img, Img, mu=mu, tukey=tukey, source_tukey=source_tukey, real=real)
    zoom = zoom_center(Z, d)
    P = zoom[zoom.shape[0] // 2, :]
    p0 = (1, np.max(zoom), 0, len(P) / 2)
    popt, pcov = scipy.optimize.curve_fit(
        fit, np.arange(len(P)), P, p0, bounds=((0, 0, -np.inf, 0), np.inf)
    )
    bg = popt[2]
    a = popt[0]
    if debug:
        return bg + np.exp(-np.abs(R) / a), Z, p0, popt
    return bg + np.exp(-np.abs(R) / a)



# Copyright 2018 Olivier Scholder <o.scholder@gmail.com>

"""
This module performs the PCS with the help of the scikit library and gives the user various function for quick plotting.
"""

import re

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib import cm
from sklearn.decomposition import PCA as PCA1
from sklearn.preprocessing import scale

from pySPM import collection

from .utils.misc import alias, aliased


@aliased
class PCA:
    def __init__(self, data):
        self.data = data
        self.pca = None
        self.standX = None

    def scatter(self, **kargs):
        pd.tools.plotting.scatter_matrix(self.data, diagonal="kde", **kargs)
        plt.tight_layout()

    def corr(self):
        corrmat = self.data.corr()
        return corrmat

    @alias("corrShow")
    def show_corr(self):
        import seaborn as sns

        corrmat = self.corr()
        sns.heatmap(corrmat).xaxis.tick_top()

    def hinton(self, max_weight=None, ax=None, matrix=None, xlabel=None, ylabel=None):
        """Draw Hinton diagram for visualizing a weight matrix."""
        if matrix is None:
            matrix = self.corr()
        ax = ax if ax is not None else plt.gca()

        if not max_weight:
            max_weight = 2 ** np.ceil(np.log(np.abs(matrix).max()) / np.log(2))

        ax.patch.set_facecolor("lightgray")
        ax.set_aspect("equal", "box")
        ax.xaxis.set_major_locator(plt.NullLocator())
        ax.yaxis.set_major_locator(plt.NullLocator())

        for (x, y), w in np.ndenumerate(matrix):
            color = "red" if w > 0 else "blue"
            size = np.sqrt(np.abs(w))
            rect = plt.Rectangle(
                [y - size / 2, x - size / 2],
                size,
                size,
                facecolor=color,
                edgecolor=color,
            )
            ax.add_patch(rect)

        nxticks = matrix.shape[1]
        nyticks = matrix.shape[0]
        if xlabel is None:
            xlabel = []
            for x in list(matrix.columns):
                x = re.sub(r"\^([0-9]+)", r"^{\1}", x)
                x = re.sub(r"_([0-9]+)", r"_{\1}", x)
                if x[-1] in ["+", "-"]:
                    x = x[:-1] + "^" + x[-1]
                xlabel.append("$" + x + "$")
        if ylabel is None:
            ylabel = list(matrix.index)
        ax.xaxis.tick_top()
        ax.set_xticks(range(nxticks))
        ax.set_xticklabels(xlabel, rotation=90)
        ax.set_yticks(range(nyticks))
        ax.set_yticklabels(ylabel)
        ax.grid(False)

        ax.autoscale_view()
        ax.invert_yaxis()

    def standardized(self, meanCentering=True):
        self.standX = pd.DataFrame(
            scale(self.data, with_mean=meanCentering),
            index=self.data.index,
            columns=self.data.columns,
        )
        return self.standX

    @alias("runPCA")
    def run_pca(self, meanCentering=True):
        if self.standX is None:
            self.standardized(meanCentering=meanCentering)
        self.pca = PCA1().fit(self.standX)

    def pca_summary(self):
        if self.pca is None:
            self.run_pca()
        names = [
            "PC" + str(i) for i in range(1, len(self.pca.explained_variance_ratio_) + 1)
        ]
        a = list(np.std(self.pca.transform(self.standX), axis=0))
        b = list(self.pca.explained_variance_ratio_)
        c = [
            np.sum(self.pca.explained_variance_ratio_[:i])
            for i in range(1, len(self.pca.explained_variance_ratio_) + 1)
        ]
        columns = pd.MultiIndex.from_tuples(
            [
                ("sdev", "Standard deviation"),
                ("varprop", "Proportion of Variance"),
                ("cumprop", "Cumulative Proportion"),
            ]
        )
        Z = zip(a, b, c)
        summary = pd.DataFrame(list(Z), index=names, columns=columns)
        return summary

    def screeplot(self, ax=None, num=None):
        if self.pca is None:
            self.run_pca()
        ax = ax if ax is not None else plt.gca()

        y = np.std(self.pca.transform(self.standX), axis=0) ** 2
        if num is None:
            num = len(y)
        x = np.arange(len(y)) + 1
        ax.grid(True)
        ax.plot(x[:num], y[:num], "o-")
        ax.set_xticks(x[:num])
        ax.set_xticklabels(["PC" + str(i) for i in x[:num]], rotation=60)
        ax.set_ylabel("Variance")

    def pc(self, id=0):
        # find the number of samples in the data set and the number of
        # variables
        if self.pca is None:
            self.run_pca()
        pc = np.matmul(self.standX.values, self.pca.components_[id].T)
        return pc

    def loadings(self, id=None):
        if self.pca is None:
            self.run_pca()
        if id is not None:
            return pd.DataFrame(
                self.pca.components_[id, None], columns=self.data.columns
            )
        return pd.DataFrame(
            self.pca.components_,
            columns=self.data.columns,
            index=[f"PC{i + 1}" for i in range(len(self.pca.components_))],
        )

    @alias("getPCAtransf")
    def get_pca_transf(self):
        if self.pca is None:
            self.run_pca()
        return self.pca.transform(self.standX)

    @alias("showStand")
    def show_stand(self):
        return pd.DataFrame(
            [self.standX.apply(np.mean), self.standX.apply(np.std)],
            index=["Mean", "Std"],
        )

    def pca_scatter(self, classifs=None, light=False):
        import seaborn as sns

        foo = self.get_pca_transf()
        if classifs is None:
            if light:
                plt.scatter(foo[:, 0], foo[:, 1])
            else:
                bar = pd.DataFrame(
                    list(zip(foo[:, 0], foo[:, 1])), columns=["PC1", "PC2"]
                )
                sns.lmplot("PC1", "PC2", bar, fit_reg=False)
        else:
            if light:
                plt.scatter(foo[:, 0], foo[:, 1], color=cm.Scalar)
            else:
                bar = pd.DataFrame(
                    list(zip(foo[:, 0], foo[:, 1], classifs)),
                    columns=["PC1", "PC2", "Class"],
                )
                sns.lmplot("PC1", "PC2", bar, hue="Class", fit_reg=False)


class ITA_PCA(PCA):
    def __init__(self, c, channels=None):
        """Init a PCA class from a collection"""
        self.col = c
        if channels is None:
            channels = c.channels.keys()
        mul = self.col.get_multivariate(channels)
        PCA.__init__(self, mul)

    @alias("showPCA")
    def show_pca(
        self,
        num=None,
        ax=None,
        pn=False,
        cmap=None,
        symmetric=True,
        loadings=True,
        **kargs,
    ):
        c = self.get_pca_col(num, pn)
        if cmap is None:
            if pn:
                cmap = "viridis"
                symmetric = False
            else:
                cmap = "bwr"
        L = self.loadings()[:num]
        if loadings:
            N = len(c)
            ncols = kargs.get("ncols", 4)
            width = kargs.get("width", 21)
            Ny = (N - 1) // ncols + 2
            Nx = min(ncols, N)
            from matplotlib.gridspec import GridSpec

            plt.figure(
                figsize=(width, (Ny - 1) * width / Nx + width * N / len(L.columns))
            )
            gs = GridSpec(
                Ny,
                Nx,
                height_ratios=[width / Nx for i in range(Ny - 1)]
                + [width * N / len(L.columns)],
            )
            ax = [plt.subplot(gs[i, j]) for i in range(Ny - 1) for j in range(Nx)]
            axh = plt.subplot(gs[-1, :])
            self.hinton(matrix=L, ax=axh)
        c.show(ax=ax, cmap=cmap, symmetric=symmetric, **kargs)
        return L

    @alias("getPCAcol")
    def get_pca_col(self, num=None, pn=False):
        if num is None:
            num = self.data.shape[1]
        assert num <= self.data.shape[1]
        PCA_col = collection.Collection(cls=self.col, name=self.col.name + "[PCA]")
        for i in range(num):
            PC = self.get_pca(i)
            if pn:
                PCA_col.add(PC * (PC >= 0), f"PC{i + 1}+")
                PCA_col.add(-PC * (PC <= 0), f"PC{i + 1}-")
            else:
                PCA_col.add(PC, f"PC{i + 1}")
        return PCA_col

    @alias("getPCA")
    def get_pca(self, id=0):
        s = next(iter(self.col.channels.values())).pixels.shape
        PC = self.pc(id).reshape(s)
        return PC



# Copyright 2018 Olivier Scholder <o.scholder@gmail.com>

"""
This module is used in order to align two different images.
Usually one from an SPM and the other from the ToF-SIMS.

This module also gives the ability to perform shift correction on images
which is used in order to align the different scans from ToF-SIMS images.
"""

import numpy as np
from scipy.ndimage import gaussian_filter
from skimage import transform as tf


class Aligner:
    def __init__(self, fixed, other, prog=True, FFT=True):
        self.fixed = np.copy(fixed)
        self.other = np.copy(other)
        self.size = fixed.shape
        self.FFT = FFT
        self.trans = [0, 0]
        self.scale = [1, 1]
        self.rotation = 0

        self.initIDX = self.getMatchingIndex()

    def compute(self, prog=False):
        # Compute the correlation to find the best shift as first guess
        if prog:
            print("Progress [1/4] Improve Shift...", end="\r")
        self.ImproveShift()
        if prog:
            print("Progress [2/6] Improve Scale X...", end="\r")
        self.ImproveScaleX()
        if prog:
            print("Progress [3/6] Improve Scale Y...", end="\r")
        self.ImproveScaleY()
        if prog:
            print("Progress [4/6] Improve Rotation...", end="\r")
        self.ImproveRotation()
        if prog:
            print("Progress [5/6] Improve Scale X...", end="\r")
        self.ImproveScaleX()
        if prog:
            print("Progress [6/6] Improve Scale Y...", end="\r")
        self.ImproveScaleY()
        if prog:
            print("", end="\r")

    def ImproveShift(self, verbose=False, **kargs):
        tform = tf.AffineTransform(
            scale=self.scale, rotation=self.rotation, translation=(0, 0)
        )

        O = tf.warp(
            self.other, tform, output_shape=self.other.shape, preserve_range=True
        )
        if self.FFT:
            Corr = np.real(
                np.fft.fftshift(
                    np.fft.ifft2(np.conj(np.fft.fft2(self.fixed)) * np.fft.fft2(O))
                )
            )
            cord = np.unravel_index(np.argmax(Corr), self.fixed.shape)
            self.trans = [cord[1] - self.size[0] / 2, cord[0] - self.size[1] / 2]
        else:
            shift, D = AutoShift(
                self.fixed, O, shift=[-self.trans[0], self.trans[1]], **kargs
            )
            if verbose:
                print("ImproveShift", shift, D)
            self.trans = [-shift[0], shift[1]]

    def getTf(self, verbose=False):
        """
        Get the Afdfine transform.
        You can apply it to a pySPM Image (img) with: img.align(this.getTf())
        """
        if verbose:
            print(f"Transpose: {self.trans[0]}, {self.trans[1]}")
        return tf.AffineTransform(
            scale=self.scale, rotation=self.rotation, translation=self.trans
        )

    def getMatchingIndex(self, power=1):
        img = tf.warp(
            self.other, self.getTf(), output_shape=self.other.shape, preserve_range=True
        )[: self.fixed.shape[0], : self.fixed.shape[1]]
        return np.sum(np.abs(self.fixed - img) ** power)

    def ImproveScaleX(self, fact=0.1, count=0, verbose=False):
        old = self.scale
        IDX1 = self.getMatchingIndex()
        self.scale[0] *= 1 + fact
        self.ImproveShift()
        IDX2 = self.getMatchingIndex()
        print("ImproveX", count, IDX1, IDX2)
        if IDX2 > IDX1:
            self.scale[0] = old[0] * (1 - fact)
            self.ImproveShift()
            IDX2 = self.getMatchingIndex()
            if IDX2 > IDX1:
                self.scale[0] = old[0]
                if fact > 1e-3:
                    self.ImproveScaleY(fact)
            elif count <= 11:
                self.ImproveScaleX(fact, count + 1)
        elif count <= 11:
            self.ImproveScaleX(fact, count + 1)

    def ImproveRotation(self, delta=0.1, count=0, prog=False):
        IDX1 = self.getMatchingIndex()
        IDX2 = IDX1
        if prog:
            print(
                f"Progress [{prog}/4] Improve Rotation. Passes: {count + 1} (max 11)",
                end="\r",
            )
        while IDX2 <= IDX1:
            IDX1 = IDX2
            count += 1
            self.rotation += delta
            self.ImproveShift()
            IDX2 = self.getMatchingIndex()
        self.rotation -= delta
        if delta > 0:
            self.ImproveRotation(-delta, count=count + 1)
        elif abs(delta) > 1e-5:
            self.ImproveRotation(abs(delta) / 10.0, count=count + 1)

    def ImproveScaleY(self, fact=0.1, count=0):
        old = self.scale
        IDX1 = self.getMatchingIndex()
        self.scale[1] *= 1 + fact
        self.ImproveShift()
        IDX2 = self.getMatchingIndex()
        if IDX2 > IDX1:
            self.scale[1] = old[1] * (1 - fact)
            self.ImproveShift()
            IDX2 = self.getMatchingIndex()
            if IDX2 > IDX1:
                self.scale[1] = old[1]
                if fact > 1e-3:
                    self.ImproveScaleX(fact / 10)
            elif count <= 11:
                self.ImproveScaleY(fact, count + 1)
        elif count <= 11:
            self.ImproveScaleY(fact, count + 1)

    def __repr__(self):
        return "Scale: ({scale[0]},{scale[1]})\nRotation: {rot:.6f} deg.\nTranslation: ({trans[0]},{trans[1]})".format(
            rot=self.rotation, scale=self.scale, trans=self.trans
        )


def ApplyShift(Img, shift):
    dx, dy = (-int(x) for x in shift)
    return np.pad(
        Img,
        ((max(0, dy), max(0, -dy)), (max(0, -dx), max(0, dx))),
        mode="constant",
        constant_values=0,
    )[max(0, -dy) : max(0, -dy) + Img.shape[0], max(0, dx) : max(0, dx) + Img.shape[1]]


def ShiftScore(
    Ref, Img, shift, gauss=5, mean=True, norm=False, debug=False, normData=False
):
    assert Ref.shape[0] <= Img.shape[0]
    assert Ref.shape[1] <= Img.shape[1]
    if mean:
        Ref = Ref - np.mean(Ref)
        Img = Img - np.mean(Img)
    if normData:
        assert np.std(Ref) > 0
        assert np.std(Img) > 0
        Ref /= np.std(Ref)
        Img /= np.std(Img)
    # blur the images for score calculation
    if gauss in [0, None, False]:
        im1 = Ref
        im2 = Img
    else:
        im1 = gaussian_filter(Ref, gauss)
        im2 = gaussian_filter(Img, gauss)

    corr2 = ApplyShift(im2, shift)
    dx, dy = shift
    # Create a copy of the reference and erase parts which are not overlaping with img
    DSX = Img.shape[1] - Ref.shape[1]
    DSY = Img.shape[0] - Ref.shape[0]
    Or = np.copy(im1)
    if dy < 0:
        Or[:-dy, :] = 0
    elif DSY - dy < 0:
        Or[DSY - dy :, :] = 0
    if dx > 0:
        Or[:, :dx] = 0
    elif DSX + dx < 0:
        Or[:, dx + DSX :] = 0

    corr2 = corr2[: Ref.shape[0], : Ref.shape[1]]
    # calculate the score: absolute of the differendces normed by the overlaping area
    D = np.sum(np.abs(Or - corr2))
    if norm:
        D /= (Ref.shape[0] - 2 * dy) * (Ref.shape[1] - 2 * dx)
    if debug:
        return D, Or, corr2
    return D


def AutoShift(
    Ref,
    Img,
    Delta=50,
    shift=(0, 0),
    step=5,
    gauss=5,
    mean=True,
    test=False,
    norm=False,
    normData=False,
):
    """Function to find the best shift between two images by using brute force
    It shift the two iumages and calculate a difference score between the two.
    The function will return the shift which gives the lowerst score (least difference)
    The score is the norm of the difference between the two images where all non-overlaping parts of the images
    due to the shifts are set to 0. The score is then normes by the effective area.
    In order to avoid the errors due to shot-noise, the images are gaussian blured.

    Delta: shifts between shift[0/1]-Delta and shift[0/1]+Delta will be tested
    step: The step between tested delta values
    gauss: For noisy image it is better to use a gaussian filter in order to improve the score accuracy.
           The value is the gaussian size.
    mean: If True, the mean value of each image are subtracted. This is prefered when the intensities of the two images does not match perfectly.
          Set it to False if you know that the intensities of your two images are identical
    Note: This function was developed as the maximum in FFT correlation does not seems to give very acurate
          result for images with low counts. If the shifts is expected to be big, the first guess shift can be calculated
          by FFT correlation. ie.:
          s = np.fft.fftshift( np.abs( np.fft.ifft2( np.fft.fft2(Reference) * np.conj(np.fft.fft2(Image)))))
          shift = [x-s.shape[i]/2 for i,x in enumerate(np.unravel_index(np.argmax(s), s.shape))]
    """
    assert Ref.shape[0] <= Img.shape[0]
    assert Ref.shape[1] <= Img.shape[1]
    if mean:
        Ref = Ref - np.mean(Ref)
        Img = Img - np.mean(Img)
    if normData:
        assert np.std(Ref) > 0
        assert np.std(Img) > 0
        Ref /= np.std(Ref)
        Img /= np.std(Img)
    # blur the images for score calculation
    if gauss in [0, None, False]:
        im1 = Ref
        im2 = Img
    else:
        im1 = gaussian_filter(Ref, gauss)
        im2 = gaussian_filter(Img, gauss)

    # the following two variables save the best score
    best = (0, 0)
    Dbest = Ref.shape[0] * Ref.shape[1] * max(np.max(im2), np.max(im1))

    tested = np.zeros((int(2 * Delta / step) + 1, int(2 * Delta / step) + 1))
    # Sweep through all possible shifts (brute force)
    for iy, Dy in enumerate(np.arange(shift[1] - Delta, shift[1] + Delta + 1, step)):
        dy = int(Dy)
        for ix, Dx in enumerate(
            np.arange(shift[0] - Delta, shift[0] + Delta + 1, step)
        ):
            dx = int(Dx)
            corr2 = ApplyShift(im2, (dx, dy))
            DSX = Img.shape[1] - Ref.shape[1]
            DSY = Img.shape[0] - Ref.shape[0]
            # Create a copy of the reference and erase parts which are not overlaping with img
            Or = np.copy(im1)
            if dy < 0:
                Or[:-dy, :] = 0
            elif DSY - dy < 0:
                Or[DSY - dy :, :] = 0
            if dx > 0:
                Or[:, :dx] = 0
            elif DSX + dx < 0:
                Or[:, dx + DSX :] = 0
            corr2 = corr2[: Ref.shape[0], : Ref.shape[1]]
            # calculate the score: absolute of the differendces normed by the
            # overlaping area
            D = np.sum(np.abs(Or - corr2))
            if norm:
                D /= (Ref.shape[0] - 2 * dy) * (Ref.shape[1] - 2 * dx)
            if test:
                tested[iy, ix] = D
            if Dbest > D:
                Dbest = D
                best = (dx, dy)
    if test:
        return best, Dbest, tested
    return best, Dbest



import array
import ctypes
import struct
from ctypes import wintypes

import psutil
import win32api
import win32con
import win32gui
from win32con import WM_GETTEXT, WM_GETTEXTLENGTH
from win32gui import PyGetString, SendMessage


def getText(hwnd):
    buffer_len = SendMessage(hwnd, WM_GETTEXTLENGTH, 0, 0) + 1
    buffer = array.array("b", b"\x00\x00" * buffer_len)
    SendMessage(hwnd, WM_GETTEXT, buffer_len, buffer)
    text = PyGetString(buffer.buffer_info()[0], buffer_len - 1)
    return text


WNDENUMPROC = ctypes.WINFUNCTYPE(wintypes.BOOL, wintypes.HWND, wintypes.LPARAM)
user32 = ctypes.windll.user32
user32.EnumWindows.argtypes = [WNDENUMPROC, wintypes.LPARAM]
user32.GetWindowTextLengthW.argtypes = [wintypes.HWND]
user32.GetWindowTextW.argtypes = [wintypes.HWND, wintypes.LPWSTR, ctypes.c_int]

prototype = ctypes.WINFUNCTYPE(
    wintypes.BOOL, wintypes.HWND, ctypes.POINTER(wintypes.RECT)
)
paramflags = (1, "hwnd"), (2, "lprect")
GetWindowRect = prototype(("GetWindowRect", user32), paramflags)


def clickButton(hwnd):
    _sendNotifyMessage(hwnd, win32con.BN_CLICKED)


def _getMultipleWindowValues(hwnd, getCountMessage, getValueMessage):
    result = []

    VALUE_LENGTH = 256

    valuecount = win32gui.SendMessage(hwnd, getCountMessage, 0, 0)
    for itemIndex in range(valuecount):
        valuebuffer = ctypes.create_unicode_buffer(VALUE_LENGTH)
        valueLength = win32gui.SendMessage(
            hwnd, getValueMessage, itemIndex, valuebuffer
        )
        result.append(valuebuffer[:valueLength])
    return result


def getListboxItems(hwnd):
    return _getMultipleWindowValues(
        hwnd, getCountMessage=win32con.LB_GETCOUNT, getValueMessage=win32con.LB_GETTEXT
    )


def findTopWindow(wantedText=None, wantedClass=None, selectionFunction=None):
    topWindows = findTopWindows(wantedText, wantedClass, selectionFunction)
    if topWindows:
        return topWindows[0]
    else:
        raise WinGuiAutoError(
            "No top level window found for wantedText="
            + repr(wantedText)
            + ", wantedClass="
            + repr(wantedClass)
            + ", selectionFunction="
            + repr(selectionFunction)
        )


def findTopWindows(wantedText=None, wantedClass=None, selectionFunction=None):
    results = []
    topWindows = []
    win32gui.EnumWindows(_windowEnumerationHandler, topWindows)
    for hwnd, windowText, windowClass in topWindows:
        if wantedText and _normaliseText(wantedText) not in _normaliseText(windowText):
            continue
        if wantedClass and windowClass != wantedClass:
            continue
        if selectionFunction and not selectionFunction(hwnd):
            continue
        results.append(hwnd)
    return results


def findWindow(name="", C="", parent=0):
    R = []

    def callbck(hwnd, lParam):
        title, c = getInfo(hwnd)
        if (name == "" or title.startswith(name)) and (C == "" or c == C):
            R.append(hwnd)
        return True

    user32.EnumChildWindows(parent, WNDENUMPROC(callbck), 42)
    return R


def listWin(hwnd, lParam):
    title, c = getInfo(hwnd)
    print(hwnd, title, c)
    return True


def getInfo(hwnd):
    length = user32.GetWindowTextLengthW(hwnd) + 1
    buffer = ctypes.create_unicode_buffer(length)
    user32.GetWindowTextW(hwnd, buffer, length)
    b2 = ctypes.create_unicode_buffer(200)
    user32.GetClassNameW(hwnd, b2, 200)
    title = buffer.value
    c = b2.value
    return title, c


def showWin(parent, level=0):
    for x in findWindow(parent=parent):
        t, c = getInfo(x)
        print("\t" * level + hex(x) + " - " + t + " - " + c)
        showWin(x, level + 1)


def getBBox(hwnd):
    b = GetWindowRect(hwnd)
    return [b.left, b.top, b.right, b.bottom]


def getPID(name):
    for proc in psutil.process_iter():
        if proc.name() == name:
            return proc.pid


def _windowEnumerationHandler(hwnd, resultList):
    """Pass to win32gui.EnumWindows() to generate list of window handle,
    window text, window class tuples."""
    resultList.append((hwnd, win32gui.GetWindowText(hwnd), win32gui.GetClassName(hwnd)))


def _normaliseText(controlText):
    """Remove '&' characters, and lower case.
    Useful for matching control text."""
    return controlText.lower().replace("&", "")


class WinGuiAutoError(Exception):
    pass


def findControl(topHwnd, wantedText=None, wantedClass=None, selectionFunction=None):
    controls = findControls(
        topHwnd,
        wantedText=wantedText,
        wantedClass=wantedClass,
        selectionFunction=selectionFunction,
    )
    if controls:
        return controls[0]
    else:
        raise WinGuiAutoError(
            "No control found for topHwnd="
            + repr(topHwnd)
            + ", wantedText="
            + repr(wantedText)
            + ", wantedClass="
            + repr(wantedClass)
            + ", selectionFunction="
            + repr(selectionFunction)
        )


def findControls(topHwnd, wantedText=None, wantedClass=None, selectionFunction=None):
    def searchChildWindows(currentHwnd):
        results = []
        childWindows = []
        try:
            win32gui.EnumChildWindows(
                currentHwnd, _windowEnumerationHandler, childWindows
            )
        except win32gui.error:
            # This seems to mean that the control *cannot* have child windows,
            # i.e. not a container.
            return
        for childHwnd, windowText, windowClass in childWindows:
            descendentMatchingHwnds = searchChildWindows(childHwnd)
            if descendentMatchingHwnds:
                results += descendentMatchingHwnds

            if wantedText and _normaliseText(wantedText) not in _normaliseText(
                windowText
            ):
                continue
            if wantedClass and windowClass != wantedClass:
                continue
            if selectionFunction and not selectionFunction(childHwnd):
                continue
            results.append(childHwnd)
        return results

    return searchChildWindows(topHwnd)


def _sendNotifyMessage(hwnd, nofifyMessage):
    """Send a notify message to a control."""
    win32gui.SendMessage(
        win32gui.GetParent(hwnd),
        win32con.WM_COMMAND,
        _buildWinLong(nofifyMessage, win32api.GetWindowLong(hwnd, win32con.GWL_ID)),
        hwnd,
    )


def _buildWinLong(high, low):
    """Build a windows long parameter from high and low words.
    See http://support.microsoft.com/support/kb/articles/q189/1/70.asp
    """
    # return ((high << 16) | low)
    return int(struct.unpack(">L", struct.pack(">2H", high, low))[0])


def selectHoriz(hwnds, i):
    controls = []
    for x in hwnds:
        bbox = getBBox(x)
        controls.append([x, *bbox])
    x0 = sorted({x[1] for x in controls})[i]
    return [x[0] for x in controls if x[1] == x0]


def selectVert(hwnds, i):
    controls = []
    for x in hwnds:
        bbox = getBBox(x)
        controls.append([x, *bbox])
    y0 = sorted({x[2] for x in controls})[i]
    return [x[0] for x in controls if x[2] == y0]


def click(x, y):
    win32api.SetCursorPos((x, y))
    win32api.mouse_event(win32con.MOUSEEVENTF_LEFTDOWN, x, y, 0, 0)
    win32api.mouse_event(win32con.MOUSEEVENTF_LEFTUP, x, y, 0, 0)



# Copyright 2018 Olivier Scholder <o.scholder@gmail.com>

import numpy as np
import scipy.optimize as opt

"""
Provides some useful mathematical functions which are not present in numpy.
"""


def prod(x):
    p = 1
    for k in x:
        p *= k
    return p


def perm(x):
    """
    permutation with repetition of n factors
    x = [n1,n2,n3,...nk]
    n = n1+n2+...+nk
    return n!/(n1!*n2*...*nk!)
    """
    n = sum(x)
    num = list(range(2, n + 1))
    denom = list(range(2, x[0] + 1))
    for k in x[1:]:
        denom += list(range(2, k + 1))
    for x in num:
        if x in denom:
            num.remove(x)
            denom.remove(x)
    return prod(num) / prod(denom)


def closest_arg(array, value):
    return np.argmin(np.abs(array - value))


def strictly_positify(x):
    """
    Make the result strictly positive by setting the minimum value to
    the lower allowed float value
    """
    return np.fmax(x, np.finfo(x.dtype).eps)


def positify(x):
    """
    Set to zero all negative values
    """
    return np.fmax(0, x)


def clip01(x):
    """
    clip data x between 0 and 1
    """
    return np.fmax(np.fmin(x, 1), 0)


def fact(x):
    """
    Return the factors of an integer as a list.
    Warning: This function is not a factorial!
    """
    if x < 0 or type(x) is not int:
        raise ValueError("input must be a positive integer")
    if x < 2:
        return x
    f = []
    i = 2
    while True:
        while x % i == 0:
            f.append(i)
            x /= i
        i += 1
        if x == 1:
            return f


def moving_average(x, N):
    assert len(x) > N
    c = np.cumsum(x)
    return (c[N:] - c[:-N]) / N


def butter_lowpass(cutOff, fs, order=5):
    import scipy.signal

    nyq = 0.5 * fs
    normalCutoff = cutOff / nyq
    b, a = scipy.signal.butter(order, normalCutoff, btype="low", analog=True)
    return b, a


def butter_lowpass_filter(data, cutOff, fs, order=4):
    import scipy.signal

    b, a = butter_lowpass(cutOff, fs, order=order)
    y = scipy.signal.lfilter(b, a, data)
    return y


def Gauss(x, x0, s, amp=None, **kargs):
    if "A" in kargs:
        from warnings import warn

        warn(
            "Parameter A is deprecated. Please use amp in order to set the amplitude!",
            stacklevel=2,
        )
        amp = kargs.pop("A")
    elif "Amp" in kargs:
        from warnings import warn

        warn(
            "Parameter Amp is deprecated. Please use amp in order to set the amplitude!",
            stacklevel=2,
        )
        amp = kargs.pop("Amp")
    R = np.exp(-((x - x0) ** 2) / (2 * s**2))
    if amp is None:
        R /= s * np.sqrt(2 * np.pi)
    else:
        R *= amp
    R[s == 0] = (x[s == 0] == x0) * 1.0
    return R


def Lorentz(x, x0, gamma, amp=None, **kargs):
    if "A" in kargs:
        from warnings import warn

        warn(
            "Parameter A is deprecated. Please use amp in order to set the amplitude!",
            stacklevel=2,
        )
        kargs["A"]
    elif "Amp" in kargs:
        from warnings import warn

        warn(
            "Parameter Amp is deprecated. Please use amp in order to set the amplitude!",
            stacklevel=2,
        )
        amp = kargs.pop("Amp")
    R = 1 / ((x - x0) ** 2 + (0.5 * gamma) ** 2)
    if amp is None:
        return 0.5 * gamma * R / np.pi
    return amp * R * (0.5 * gamma) ** 2


def CDF(x, mu, sig, amp=1, lg=0, **kargs):
    if "Amp" in kargs:
        from warnings import warn

        warn(
            "Parameter Amp is deprecated. Please use amp in order to set the amplitude!",
            stacklevel=2,
        )
        amp = kargs.pop("Amp")
    from scipy.special import erf

    g = sig * np.sqrt(2 * np.log(2))
    return amp * lg * (0.5 + np.arctan2(x - mu, g) / np.pi) + (1 - lg) * amp * 0.5 * (
        1 + erf((x - mu) / (sig * np.sqrt(2)))
    )


def LG(x, x0, sig=None, amp=None, lg=0.5, asym=1, FWHM=None, **kargs):
    if "Amp" in kargs:
        from warnings import warn

        warn(
            "Parameter Amp is deprecated. Please use amp in order to set the amplitude!",
            stacklevel=2,
        )
        amp = kargs.pop("Amp")

    assert sig is not None or FWHM is not None

    if FWHM is None:
        FWHM = 2 * np.sqrt(2 * np.log(2)) * sig
    if sig is None:
        sig = FWHM / (2 * np.sqrt(2 * np.log(2)))
    Y = (1 - lg) * Gauss(x, x0, sig, amp=amp) + lg * Lorentz(x, x0, FWHM, amp=amp)
    if asym != 1:
        Yr = (1 - lg) * Gauss(x, x0, sig * asym, amp=amp) + lg * Lorentz(
            x, x0, FWHM * asym, amp=amp
        )
        Y[x > x0] = Yr[x > x0]
    return Y


def logistic(x, lower=0, upper=1, growth=1, x0=0, nu=1, C=1):
    return lower + (upper - lower) / (C + np.exp(-growth * (x - x0))) ** (1 / nu)


def fitCDF1line(A):
    line = np.zeros(A.shape[1])
    for x in range(A.shape[1]):
        popt, pcov = opt.curve_fit(
            CDF,
            np.arange(A.shape[0]),
            A[:, x],
            (A.shape[0] / 2, 1, np.max(A[:, x])),
            bounds=(0, (A.shape[1], np.inf, np.inf)),
        )
        line[x] = popt[0]
    return line


def FT(x, ufunc=np.real, real=False):
    """
    shortcut for 1D/2D Fourier Transform (real and centered)
    """
    assert isinstance(x, np.ndarray)
    if len(x.shape) == 1:
        F = np.fft.rfft(x) if real else np.fft.fft(x)
    elif len(x.shape) == 2:
        F = np.fft.rfft2(x) if real else np.fft.fft2(x)
    else:
        raise TypeError("The array should be 1D or 2D")
    return ufunc(np.fft.fftshift(F))


def binning(data, N=2, axis=0, ufunc=np.sum):
    w = int(np.floor(data.shape[axis] / N))
    r = np.copy(data)
    size = list(data.shape)
    size[axis] = w
    size = size[: axis + 1] + [N] + size[axis + 1 :]
    r.resize(size)
    return ufunc(r, axis=axis + 1)


def stat_info(data):
    import matplotlib.pyplot as plt

    D = np.ravel(data)
    U = np.unique(D)
    if len(U) > 1:
        sep = np.min(U[1:] - U[:-1])
        N = min(100, int(np.ceil((np.max(D) - np.min(D)) / sep)))
    else:
        N = 1

    mean = np.mean(D)
    std = np.std(D)

    fig, ax = plt.subplots(2, 1, figsize=(21, 4))
    ax[0].boxplot(D, 0, "ro", 0)
    ax[1].hist(D, N, density=True)
    ax[1].axvline(mean, color="r", label="mean")
    ax[1].axvline(mean + std, color="r", linestyle="--", label="1$\\sigma$")
    ax[1].axvline(mean - std, color="r", linestyle="--", label="1$\\sigma$")
    if mean - 2 * std >= U[0]:
        ax[1].axvline(mean - 2 * std, color="r", linestyle=":", label="2$\\sigma$")
    if mean + 2 * std <= U[-1]:
        ax[1].axvline(mean + 2 * std, color="r", linestyle=":", label="2$\\sigma$")
    ax[1].legend()
    print("Stats")
    print("\tAverage:", mean)
    print("\tStandard-deviation:", std)
    print("\tMinimum:", np.min(D))
    print("\tQ1:", np.percentile(D, 25))
    print("\tMedian:", np.percentile(D, 50))
    print("\tQ3:", np.percentile(D, 75))
    print("\tMaximum:", np.max(D))


def ellipse(a, b, phi):
    """
    return the x,y coordinates of an ellipse with major axis=a and minor axis=b at angle phi(radians)
    """
    r = np.zeros(phi.shape)  # if a & b is zero, result is zero
    m = np.logical_and(a != 0, b != 0)
    r[m] = a * b / np.sqrt((b * np.cos(phi[m])) ** 2 + (a * np.sin(phi[m])) ** 2)
    return r


def asymm_ellipse(left, right, upper, lower, phi):
    phi = np.divmod(phi + 2 * np.pi, 2 * np.pi)[1]  # Be sure phi ∈ [0,2π]
    b = np.where(phi <= np.pi, upper, lower)
    a = np.where(np.logical_or(phi <= np.pi / 2, phi >= 3 * np.pi / 2), right, left)
    m = np.logical_and(a != 0, b != 0)
    r = np.zeros(phi.shape)  # if a & b is zero, result is zero
    r[m] = (
        a[m]
        * b[m]
        / np.sqrt((b[m] * np.cos(phi[m])) ** 2 + (a[m] * np.sin(phi[m])) ** 2)
    )
    return r


def LG2D(
    XY,
    amplitude=1,
    angle=0,
    sig_x=10,
    sig_y=10,
    x0=None,
    y0=None,
    LG_x=0,
    LG_y=0,
    assym_x=1,
    assym_y=1,
    bg=0,
):
    """
    Return a 2D Lorentz-Gauss.
    XY: (X,Y) tuple
    A: amplitude
    a: angle
    sx: sigma for x-axis
    sy: sigma fdor y-axis
    x0,y0 : center coordinates of the peak
    lgx, lgy: Lorentz-Gauss proportion (for x,y axis)
    assym_x, assym_y: The assymetry in sig_x/sig_y for the left/right or upper/lower part of the curve
    """
    if x0 is None:
        x0 = XY[0][0, XY[0].shape[1] // 2]
    if y0 is None:
        y0 = XY[1][XY[1].shape[0] // 2, 0]
    X1 = (XY[0] - x0) * np.cos(angle) - (XY[1] - y0) * np.sin(angle)
    Y1 = (XY[0] - x0) * np.sin(angle) + (XY[1] - y0) * np.cos(angle)

    R1 = np.sqrt(X1**2 + Y1**2)
    angle = np.arctan2(Y1, X1)
    sig = asymm_ellipse(sig_x, sig_x * assym_x, sig_y, sig_y * assym_y, angle)
    gamma = np.sqrt(2 * np.log(2)) * sig
    LG = ellipse(LG_x + 1, LG_y + 1, angle) - 1

    Gxy = Gauss(R1, 0, sig, 1)
    Lxy = 1 / ((R1 / gamma) ** 2 + 1)

    f = (1 - LG) * Gxy + LG * Lxy
    out = bg + amplitude * f
    return out


def LG2Da(
    XY,
    amplitude=1,
    angle=0,
    sigN=10,
    sigS=None,
    sigE=10,
    sigW=None,
    x0=None,
    y0=None,
    LGN=0,
    LGS=None,
    LGE=0,
    LGW=None,
    bg=0,
):
    if x0 is None:
        x0 = XY[0][0, XY[0].shape[1] // 2]
    if y0 is None:
        y0 = XY[1][XY[1].shape[0] // 2, 0]
    if sigS is None:
        sigS = sigN
    if sigW is None:
        sigW = sigE
    if LGS is None:
        LGS = LGN
    if LGW is None:
        LGW = LGE

    X1 = (XY[0] - x0) * np.cos(angle) - (XY[1] - y0) * np.sin(angle)
    Y1 = (XY[0] - x0) * np.sin(angle) + (XY[1] - y0) * np.cos(angle)

    R1 = np.sqrt(X1**2 + Y1**2)
    angle = np.arctan2(Y1, X1)
    sig = asymm_ellipse(sigW, sigE, sigN, sigS, angle)
    gamma = np.sqrt(2 * np.log(2)) * sig  # HFHM
    LG = asymm_ellipse(LGW + 1, LGE + 1, LGN + 1, LGS + 1, angle) - 1

    Gxy = Gauss(R1, 0, sig, 1)
    Lxy = 1 / ((R1 / gamma) ** 2 + 1)

    f = (1 - LG) * Gxy + LG * Lxy
    out = bg + amplitude * f
    return out


def MaxwellBoltzmann(E, T):
    from . import constants as const

    return (
        2
        * const.qe
        * np.sqrt(E / np.pi)
        * np.exp(-E / (const.kb * T))
        / (const.kb * T) ** 1.5
    )


def Voigt(x, x0, sig, gamma, A=1):
    import scipy

    L = Lorentz(x, x0, gamma, A=1)
    G = Gauss(x, x0, sig)
    out = scipy.signal.convolve(L, G, "same")
    out /= np.max(out)
    return A * out



import sys

import matplotlib as mpl
import matplotlib.pyplot as plt
import pandas as pd

from pySPM.tools.fpanel import Fpanel


def plotLog(filename, watch=False, **kargs):
    fig, ax = plt.subplots(1, 1)
    fig.subplots_adjust(hspace=0)
    plt.show(block=False)
    while True:
        with open(filename) as f:
            names = f.readline().rstrip().split("\t")
        df = pd.read_csv(
            filename,
            skiprows=1,
            delimiter="\t",
            parse_dates=[0],
            na_values="<undefined>",
            names=names,
        )
        # df = df.dropna()
        ax2 = df.plot("Time", subplots=True, ax=ax, sharex=True)
        dt = df.iloc[-1, 0] - df.iloc[0, 0]
        for a in ax2:
            if dt.seconds < 15 * 60:
                a.xaxis.set_major_locator(mpl.dates.MinuteLocator(interval=1))
            elif dt.seconds < 3 * 60 * 60:
                a.xaxis.set_major_locator(mpl.dates.MinuteLocator(interval=5))
            else:
                a.xaxis.set_major_locator(mpl.dates.MinuteLocator(interval=15))
            a.xaxis.set_major_formatter(mpl.dates.DateFormatter("%H:%M"))
            a.grid()
        plt.minorticks_off()

        if watch:
            mypause(3)
        else:
            plt.show()
        if not watch:
            break


def mypause(interval):
    backend = plt.rcParams["backend"]
    if backend in mpl.rcsetup.interactive_bk:
        figManager = mpl._pylab_helpers.Gcf.get_active()
        if figManager is not None:
            canvas = figManager.canvas
            if canvas.figure.stale:
                canvas.draw()
            canvas.start_event_loop(interval)
            return


def main():
    if len(sys.argv) > 1:
        filename = sys.argv[1]
        print(f'Plot file "{filename}"')
        plotLog(filename, watch=False)
    else:
        F = Fpanel()
        logfile = F.getLogFile()
        plotLog(logfile, watch=True)


if __name__ == "__main__":
    main()



# Copyright 2018 Olivier Scholder <o.scholder@gmail.com>

import base64
import os
import struct
import xml.etree.ElementTree as ElementTree

import numpy as np

import pySPM.SPM
from pySPM.SPM import SPM_image, funit

from .utils.misc import alias, aliased, deprecated


@deprecated("getCurve")
def get_curve(filename, channel="Normal Deflection", backward=False):
    """
    function to retrieve data which are not in the form of images.
    This is typically used for 1D channel where the normal deflection is recorded while z is swept.
    """
    tree = ElementTree.parse(filename)
    root = tree.getroot()
    namespace = {"spm": "http://www.nanoscan.ch/SPM"}
    RAW = root.findall(
        "spm:vector/spm:contents/spm:direction/spm:vector/"
        "spm:contents/spm:name[spm:v='{direction}']/../spm:channel/spm:vector/"
        "spm:contents/spm:name[spm:v='{channel}']/../spm:data/spm:v".format(
            direction=["forward", "backward"][backward], channel=channel
        ),
        namespace,
    )[0].text
    start = float(
        root.findall(
            "spm:vector/spm:contents/spm:axis/spm:vector/"
            "spm:contents/spm:start/spm:vector/spm:v",
            namespace,
        )[0].text
    )
    stop = float(
        root.findall(
            "spm:vector/spm:contents/spm:axis/spm:vector/"
            "spm:contents/spm:stop/spm:vector/spm:v",
            namespace,
        )[0].text
    )
    root.findall(
        "spm:vector/spm:contents/spm:axis/spm:vector/spm:contents" "/spm:unit/spm:v",
        namespace,
    )[0].text
    BIN = base64.b64decode(RAW)
    N = len(BIN)
    vals = np.array(struct.unpack("<" + str(N // 4) + "f", BIN))
    x = np.linspace(start, stop, len(vals))
    return x, vals


@aliased
class Nanoscan:
    def __init__(self, filename=None):
        if not os.path.exists(filename):
            raise OSError(f'File "{filename}" Not Found')
        if filename[-4:] != ".xml":
            raise TypeError("Nanoscan files should be xml files")
        self.filename = filename
        tree = ElementTree.parse(filename)
        self.root = tree.getroot()

        if self.root.tag == "{http://www.nanoscan.ch/SPM}scan":
            self.namespaces = {"spm": "http://www.nanoscan.ch/SPM"}
            self.type = "Nanoscan"
            self.fbPath = "spm:vector/spm:contents/spm:instrumental_parameters/spm:contents/spm:z_control/spm:contents"
            self.pixel_size = [
                int(z)
                for z in self.__grab(
                    "spm:vector/spm:contents/spm:size/spm:contents//spm:v"
                )
            ]
            uval = float(
                self.__grab(
                    ".//spm:area//spm:contents/spm:size/spm:contents/spm:fast_axis/spm:v"
                )
            )
            udispu = self.__grab(".//spm:area//spm:contents/spm:display_unit/spm:v")
            udisps = float(
                self.__grab(".//spm:area/spm:contents/spm:display_scale/spm:v")
            )
            self.__grab(".//spm:area/spm:contents/spm:unit/spm:v")
            x = funit(uval * udisps, udispu)
            uval = float(
                self.__grab(
                    ".//spm:area//spm:contents/spm:size/spm:contents/spm:slow_axis/spm:v"
                )
            )
            y = funit(uval * udisps, udispu)
            self.size = {"unit": x["unit"], "x": x["value"], "y": y["value"]}
            try:
                self.feedback = {
                    "channel": self.__grab(
                        f"{self.fbPath}/spm:z_feedback_channel/spm:v"
                    )
                }
                self.feedback["P"] = {
                    "value": float(
                        self.__grab(f"{self.fbPath}/spm:proportional_z_gain/spm:v")
                    ),
                    "unit": self.__grab(
                        f"{self.fbPath}/spm:proportional_z_gain_unit/spm:v"
                    ),
                }
                self.feedback["I"] = {
                    "value": float(
                        self.__grab(f"{self.fbPath}/spm:integral_z_time/spm:v")
                    ),
                    "unit": self.__grab(
                        f"{self.fbPath}/spm:integral_z_time_unit/spm:v"
                    ),
                }
                if self.feedback["channel"] == "df":
                    self.feedback["channel"] = "Δf"
            except:
                self.feedback = {}
            self.scan_speed = {
                z: {
                    "value": float(
                        self.__grab(
                            f"spm:vector//spm:direction/spm:vector/spm:contents/spm:name[spm:v='{z}']/../spm:point_interval/spm:v"
                        )
                    )
                    * self.pixel_size[0],
                    "unit": self.__grab(
                        f"spm:vector//spm:direction/spm:vector/spm:contents/spm:name[spm:v='{z}']/../spm:point_interval_unit/spm:v"
                    ),
                }
                for z in ["forward", "backward"]
            }
        else:
            raise TypeError(
                "Unknown or wrong data type. Expecting a valid Nanoscan xml"
            )

    def list_channels(self):
        """
        Printout the list of stored channels
        """
        for d in ["forward", "backward"]:
            print(d)
            print("=" * len(d))
            for z in self.root.findall(
                f"spm:vector//spm:direction/spm:vector/spm:contents/spm:name[spm:v='{d}']/../spm:channel//spm:contents/spm:name/spm:v",
                self.namespaces,
            ):
                print("  - " + z.text)
            print()

    def get_channel(self, channel="Topography", backward=False, corr=None):
        try:
            RAW = self.__grab(
                "spm:vector//spm:direction/spm:vector/spm:contents"
                "/spm:name[spm:v='{direction}']/../spm:channel//spm:contents/spm:name[spm:v='{channel}']"
                "/../spm:data/spm:v".format(
                    direction=["forward", "backward"][backward], channel=channel
                )
            )
        except:
            raise "Channel {} in {} scan not found".format(
                channel, "backward" if backward else "forward"
            )

        BIN = base64.b64decode(RAW)
        recorded_length = len(BIN) / 4

        py = int(recorded_length / self.pixel_size[0])
        recorded_size = {
            "x": self.size["x"],
            "y": self.size["y"] * py / float(self.pixel_size[1]),
            "unit": self.size["unit"],
        }

        image_array = np.array(struct.unpack("<%if" % (recorded_length), BIN)).reshape(
            (py, self.pixel_size[0])
        )
        return SPM_image(
            image_array, channel=channel, _type=self.type, real=recorded_size, corr=corr
        )

    def __grab(self, path):
        result = [z.text for z in self.root.findall(path, self.namespaces)]
        if len(result) == 1:
            result = result[0]
        return result

    @deprecated("arraySummary")
    def array_summary(self):
        [
            y.format(**self.__dict__)
            for y in [
                "{filename}",
                "{pixel_size[0]}×{pixel_size[1]}",
                "{size[x][value]}×{size[y][value]} {size[x][unit]}",
                "{scan_speed[forward][value]} {scan_speed[forward][unit]}",
                "{feedback[channel]}",
                "{P[value]:.2f} {P[unit]}",
                "{I[value]:.2f} {I[unit]}",
            ]
        ]

    @alias("getSummary")
    def get_summary(self):
        x = funit(self.size["x"], self.size["unit"])
        y = funit(self.size["y"], self.size["unit"])
        P = funit(self.feedback["P"])
        I = funit(self.feedback["I"])
        return """Feedback: {feedback[channel]} : P:{P[value]}{P[unit]} : I:{I[value]}{I[unit]}
Size: {pixel_size[0]}×{pixel_size[1]} pixels = {x[value]:.3} {x[unit]}×{y[value]:.3} {y[unit]}
Scan Speed: {scanSpeed[value]}{scanSpeed[unit]}/line""".format(
            x=x,
            y=y,
            P=P,
            I=I,
            feedback=self.feedback,
            pixel_size=self.pixel_size,
            scanSpeed=self.scan_speed["forward"],
        )

    @staticmethod
    def show_dir_summary(path):
        from pySPM.utils import htmlTable

        res = [
            ["Filename", "pixel size", "real size", "scan_speed", "feedback", "P", "I"]
        ]
        for x in os.listdir(path):
            try:
                A = pySPM.Nanoscan(path + x)
                res.append(
                    [
                        y.format(f=os.path.basename(A.filename), **A.__dict__)
                        for y in [
                            "{f}",
                            "{pixel_size[0]}×{pixel_size[1]}",
                            "{size[x]}×{size[y]} {size[unit]}",
                            "{scan_speed[forward][value]} {scan_speed[forward][unit]}",
                            "{feedback[channel]}",
                            "{feedback[P][value]:.2f} {feedback[P][unit]}",
                            "{feedback[I][value]:.2f} {feedback[I][unit]}",
                        ]
                    ]
                )
            except:
                print('Cannot read image "' + x + '" skipping it')
        htmlTable(res, header=True)



import functools
from warnings import warn

import numpy as np


def dec_debug(debug):
    if debug > 0:
        return debug - 1
    if debug < 0:
        return debug + 1
    return False


def do_debug(debug):
    return debug == -1 or debug > 0


def adaptive(img):
    import skimage

    mi = np.min(img)
    img = np.asarray(256**2 * (img - mi) / (np.max(img) - mi), dtype=np.uint16)
    img = skimage.exposure.equalize_adapthist(img, clip_limit=0.03)
    return img


def smiley(
    width,
    height=None,
    ratio=0.9,
    eye=0.15,
    thick=0.1,
    eye_sep=0.35,
    eye_height=0.3,
    mouth_rad=0.5,
    mouth_thick=0.1,
):
    if height is None:
        height = width
    size = min(width, height)
    Y, X = np.mgrid[-height // 2 : height // 2, -width // 2 : width // 2]
    R = np.sqrt(X**2 + Y**2) / (size / 2)
    smiley = (ratio > R) * ((ratio - thick) < R)
    smiley += (
        np.sqrt(
            (X - eye_sep * ratio * size / 2) ** 2
            + (Y + eye_height * ratio * size / 2) ** 2
        )
        < ratio * eye * size / 2
    )
    smiley += (
        np.sqrt(
            (X + eye_sep * ratio * size / 2) ** 2
            + (Y + eye_height * ratio * size / 2) ** 2
        )
        < ratio * eye * size / 2
    )
    smiley += (
        (ratio * mouth_rad > R)
        * ((ratio * mouth_rad - ratio * mouth_thick) < R)
        * (Y > 0)
    )
    return smiley * 1.0


def aliased(cls):
    original_methods = cls.__dict__.copy()
    for name in original_methods:
        method = original_methods[name]
        if hasattr(method, "_aliases"):
            for alias in method._aliases:
                setattr(cls, *alias)
    return cls


class alias:
    def __init__(self, *aliases):
        self.aliases = aliases

    def __call__(self, func):
        if not hasattr(func, "_aliases"):
            func._aliases = []
        func._aliases += [(x, func) for x in self.aliases]
        import sys

        module = sys.modules[func.__module__]
        for alias in self.aliases:
            setattr(module, alias, func)
        return func


class deprecated:
    def __init__(self, alias):
        self.alias = alias

    def __call__(self, func):
        import sys

        msg = "the function {} is DEPRECATED. Please use {} instead".format(
            self.alias, func.__name__
        )
        module = sys.modules[func.__module__]

        @functools.wraps(func)
        def wrapper(*args, **kargs):
            from warnings import warn

            warn(msg, stacklevel=2)
            return func(*args, **kargs)

        setattr(module, self.alias, wrapper)
        if not hasattr(func, "_aliases"):
            func._aliases = []
        func._aliases += [(self.alias, wrapper)]
        return func


@alias("get_bam", "get_BAM")
def getBAM(x, x0, N=10, least_one=False):
    """
    get the profile of the GaAlAs layer from a BAM-L200 sample.

    Parameters
    ----------
    x: numpy.ndarray
        x-axis in nm
    x0: float
        offset of the x-axis in nm (ie. where the first edge of the BAM is located)
    N: int
        The number of headings grating (seems to be 13 on the SEM image in the doc, but 10 in reality?)
    least_one: bool
        if True each GaAlAs strip are at least 1px wide.
    """
    P = x * 0

    # Here are the certified known distances
    d = {
        "W1": 691,
        "W2": 691,
        "W3": 293,
        "W4": 294,
        "W5": 19.5,
        "W6": 195,
        "W7": 195,
        "W8": 38,
        "W9": 3.6,
        "W10": 14.2,
        "W11": 3.5,
        "W12": 96,
        "W13": 5,
        "W14": 1,
        "P1": 587,
        "P2": 389,
        "P3": 273,
        "P4": 193,
        "P5": 136,
        "P6": 97,
        "P7": 67.5,
        "P8": 48.5,
        "P9": 76.5,
        "P10": 57,
        "P11": 42,
        "P12": 31,
        "P13": 23,
        "P14": 17.5,
        "P15": 13.3,
        "P16": 9.4,
        "P17": 6.9,
        "P18": 4.6,
        "P19": 3,
        "P20": 2,
    }
    # The prelines
    lines = [(-147 * i, 80) for i in range(N, 0, -1)]

    # Estimated coordinates of the starting of lines
    pos = {
        "W1": 0,
        "P1": d["W1"] + d["W2"],
        "W5": 2683,
        "P9": 8760,
        "W8": 7300,
        "W9": 7800,
        "W10": 8250,
        "W11": 9500,
        "W12": 9840,
        "W14": 8100,
    }
    pos_rel = {
        "W5": {
            "P2": 397,
            "P3": 1380,
            "P4": 2140,
            "P5": 2730,
            "P6": 3200,
            "P7": 3535,
            "P8": 3730,
        },
        "P9": {"P10": 231, "P11": 402, "P12": 539, "P13": 642, "P14": 710},
        "W11": {"P15": 60, "P16": 104, "P17": 141, "P18": 162, "P19": 182, "P20": 191},
    }
    for rel in pos_rel:
        for key in pos_rel[rel]:
            pos[key] = pos[rel] + pos_rel[rel][key]

    for key in sorted(pos):
        width = [d[key], d[key] // 2][key[0] == "P"]
        lines.append((pos[key], width))
        if key[0] == "P":
            lines.append((pos[key] + d[key], width))

    for pos, width in lines:
        P[((x - x0) >= pos) * ((x - x0) <= pos + width)] = 1
        if least_one:
            P[np.argmin(abs(x - x0 - pos))] = 1  # At least 1 pixel set ?
    return P


def in_ipynb():
    try:
        # noinspection PyUnresolvedReferences
        cfg = get_ipython().config
        return "IPKernelApp" in cfg
    except NameError:
        return False


if in_ipynb():
    try:
        from tqdm.notebook import tqdm
    except:
        try:
            from tqdm import tqdm
        except:
            warn(
                "the library tqdm cannot be found. All progressbar will be disabled.",
                stacklevel=2,
            )

            def tqdm(x):
                return x

    PB = tqdm
else:
    try:
        from tqdm import tqdm
    except:
        warn(
            "the library tqdm cannot be found. All progressbar will be disabled.",
            stacklevel=2,
        )

        def tqdm(x):
            return x

    PB = tqdm



import numpy as np
import pywt


def sign(abs_var, sign_var):
    return abs(abs_var) * (1 - np.where(sign_var < 0, 2 * sign_var, sign_var))


def hfilter(diff_image, var_image, threshold=1, ndamp=10):
    """
    This code was inspired from: https://github.com/spacetelescope/sprint_notebooks/blob/master/lucy_damped_haar.ipynb
    I believe it was initially written by Justin Ely: https://github.com/justincely
    It was buggy and not working properly with every image sizes.
    I have thus exchanged it by using pyWavelet (pywt) and a custom function htrans
    to calculate the matrix for the var_image.
    """
    him, coeff_slices = pywt.coeffs_to_array(
        pywt.wavedec2(diff_image.astype(np.float), "haar"), padding=0
    )
    dvarim = htrans(var_image.astype(np.float))

    sqhim = ((him / threshold) ** 2) / dvarim
    index = np.where(sqhim < 1)

    if len(index[0]) == 0:
        return diff_image

    # Eq. 8 of White is derived leading to N*x^(N-1)-(N-1)*x^N  :DOI: 10.1117/12.176819
    sqhim = sqhim[index] * (
        ndamp * sqhim[index] ** (ndamp - 1) - (ndamp - 1) * sqhim[index] ** ndamp
    )
    him[index] = sign(threshold * np.sqrt(dvarim[index] * sqhim), him[index])

    return pywt.waverec2(
        pywt.array_to_coeffs(him, coeff_slices, output_format="wavedec2"), "haar"
    )[: diff_image.shape[0], : diff_image.shape[1]]


def htrans(A):
    h0 = A
    res = []
    while h0.shape[0] > 1 and h0.shape[1] > 1:
        h0, (hx, hy, hc) = pywt.dwt2(h0, "haar")
        res = [(h0, h0, h0), *res]
    out, _ = pywt.coeffs_to_array([h0, *res], padding=1)
    return out



# Copyright 2018 Olivier Scholder <o.scholder@gmail.com>

from .constants import *
from .elts import *
from .math import *
from .misc import PB, alias
from .plot import *
from .restoration import *
from .save import *
from .spectra import *


def funit(value, unit=None):
    """
    Convert a value and unit to proper value
    e.g. 0.01m will be converted to 10mm
    e.g. 1000A will be converted to 1kA
    etc.

    Parameters
    ----------
    value : int, float or dictionary with 'value' and 'unit' key
        numerical value to work with
    unit : string or None
        name of the unit

    Returns
    -------
    dictionary with formated 'value' and 'unit'.
    The value will be: ≥1 and <1000

    Examples
    --------
    >>> funit(0.01,'m')
    {'value': 10.0, 'unit': 'mm'}
    >>> funit(1, 'cm')
    {'value': 1.0, 'unit': 'cm'}
    >>> funit({'value':2340, 'unit': 'um'})
    {'value': 2.34, 'unit': 'mm'}
    """
    if unit is None:
        unit = value["unit"]
        value = value["value"]
    import math

    if value == 0:
        return {"value": value, "unit": unit}
    shift = int(math.floor(math.log10(value) / 3.0))  # base10 exponent
    mag = "afpnum1kMGTPE"
    index_mag = mag.index("1")
    # Test if unit has a scale factor (n,m,k,M,G, etc.)
    if len(unit) > 1 and unit[0] in mag and unit[1:] and index_mag:
        index_mag = mag.index(unit[0])
        unit = unit[1:]
    value /= 10 ** (3 * shift)
    index_mag += shift
    if index_mag < 0:
        value *= 10 ** (index_mag * 3)
        index_mag = 0
    elif index_mag >= len(mag):
        value *= 10 ** ((index_mag - len(mag) + 1) * 3)
        index_mag = len(mag) - 1
    unit_prefix = mag[index_mag]
    if unit_prefix == "1":
        unit_prefix = ""
    return {"value": value, "unit": f"{unit_prefix}{unit}"}


def get_shifts_bbox(shifts, shape):
    """
    Return the bounding-box of the overlapping area of scans which have a thermal drift given by shifts.
    Return: dictionary with the top, left, right, bottom position

    Parameters
    ----------
    shifts: list of tuple [(dx0, dy0), (dx1, dy1), ..., (dxN, dyN)]
        list of tuple given the shift for each scan in pixel in the x- and y-direction.
        The number of element should equal the number of scans (so N=Nscan-1)
    shape: tuple (Height, Width)
        shape of the image.
    """
    from .geometry import Bbox

    right = shape[1] + min(0, np.min([x[0] for x in shifts]))
    left = np.max([x[0] for x in shifts])
    top = shape[0] + min(0, np.min([x[1] for x in shifts]))
    bottom = np.max([x[1] for x in shifts])
    return Bbox(right=right, left=left, top=top, bottom=bottom)


def s2hms(s):
    """Convert seconds to hour, minutes or seconds"""
    M = np.max(s)
    if M > 120 * 60:
        return s / (60 * 60), "h"
    if M > 300:
        return s / 60, "min"
    return s, "s"


def time2hms(s, string=True):
    h = int(s // 3600)
    s -= h * 3600
    m = int(s // 60)
    s -= m * 60
    if string:
        return f"{h:02d}:{m:02d}:{s:02.2f}"
    return (h, m, s)


@alias("fitSpectrum")
def fit_spectrum(t, m, error=False):
    """
    fit and return the sf and k0 parameters for given known times t and masses m
    """
    M = np.sqrt(np.array(m))
    T = np.hstack([np.array(t)[:, None], np.ones((len(t), 1))])
    x = np.linalg.lstsq(T, M, rcond=-1)[0]
    sf = 1 / x[0]
    k0 = -sf * x[1]
    res = [sf, k0]
    if error:
        mfit = np.dot(T, x)
        mresid = M - mfit
        X = np.dot(T.T, T)
        epsvar = np.var(mresid, ddof=2)
        xvar = np.linalg.inv(X) * epsvar  # correlation matrix
        D = np.sqrt(np.diag(xvar))  # error vector
        Dsf = np.abs(D[0] / x[0] ** 2)
        Dk0 = np.sqrt((D[1] / x[0]) ** 2 + (x[1] * Dsf) ** 2)
        res += [Dsf, Dk0]
    return res


def chunks(l, n):
    """Yield successive n-sized chunks from l."""
    for i in range(0, len(l), n):
        yield l[i : i + n]


def mass2time(m, sf, k0):
    """
    Here the time is actually the number of channels.
    To convert it to real time, you should multiply the answer by the Time Resolution which can be obtained by pySPM.ITM.get_value("Registration.TimeResolution")
    """
    r = m * 0 + k0
    if not hasattr(r, "__setitem__"):
        if m < 0:
            return r
        return sf * np.sqrt(m) + k0
    r[m >= 0] = sf * np.sqrt(m[m >= 0]) + k0
    return r


def time2mass(t, sf, k0):
    """
    Here the time t is actually the number of channels.
    To convert a real time to the number of channel t, you should divide the real time by the Time Resolution which can be obtained by pySPM.ITM.get_value("Registration.TimeResolution")
    """
    return ((t - k0) / sf) ** 2


def show_table(t):
    from IPython.core.display import HTML, display

    display(HTML(html_table(t)))


def html_table(t, header=False):
    S = "<table>"
    if header:
        S += "<tr>" + "".join([f"<th>{x}</th>" for x in t[0]]) + "</tr>"
        t = t[1:]
    for x in t:
        S += "<tr>" + "".join([f"<td>{y}</td>" for y in x]) + "</tr>"
    S += "</table>"
    return S


def aa_table(t, header=False):
    """
    print a list of list in a nice ascii-art
    """
    Ncols = len(t[0])
    Lcol = [0] * Ncols
    for x in t:
        for i in range(Ncols):
            Lcol[i] = max(Lcol[i], len(repr(x[i])))
    if header:
        print(
            "  ".join(["{: <" + str(Lcol[i] + 4) + "}" for i in range(Ncols)]).format(
                *t[0]
            )
        )
        print("=" * sum(Lcol))
        t = t[1:]
    for j, x in enumerate(t):
        print(
            "  ".join(
                [
                    "{:" + [".", "_"][j % 2] + "<" + str(Lcol[i] + 4) + "}"
                    for i in range(Ncols)
                ]
            ).format(*x)
        )


def dict_update(d, u):
    import collections

    for k, v in u.items():
        if isinstance(v, collections.Mapping):
            r = dict_update(d.get(k, {}), v)
            d[k] = r
        else:
            d[k] = u[k]
    return d


def htmlTable(t, show=True, header=False):
    import html

    s = "<table><tr>"
    if header:
        s += (
            "<th>"
            + "</th><th>".join([html.escape(str(y)) for y in t[0]])
            + "</th></tr><tr>"
        )
        t = t[1:]
    s += (
        "".join(
            [
                "<tr><td>"
                + "</td><td>".join([html.escape(str(y)) for y in x])
                + "</td></tr>"
                for x in t
            ]
        )
        + "</table>"
    )
    if show:
        from IPython.display import HTML, display

        display(HTML(s))
    else:
        return s


def zfit_level(A, processes=4):
    import multiprocessing as mp

    level = np.array(
        mp.Pool(processes).map(fitCDF1line, (A[:, i, :] for i in range(A.shape[1])))
    )
    return level


def getToFimg(I, N=100, prog=False):
    """
    Simulate obtained counts for N scans from an image
    """
    R = np.zeros(I.shape)
    L = range(N)
    if prog:
        L = PB(L)
    for _i in L:
        R += np.random.rand(*I.shape) < (1 - np.exp(-I))
    return R


def getToFsimg(I, N=None, prog=False):
    if N is None:
        N = [10, 50, 100, 300, 500]
    Ns = np.insert(np.diff(N), 0, N[0])
    T = [getToFimg(I, n) for n in Ns]
    return dict(zip(N, np.cumsum(T, axis=0)))


def centered_meshgrid(A):
    w = A.shape[1]
    h = A.shape[0]
    Y, X = np.mgrid[-h // 2 : h // 2, -w // 2 : w // 2]
    return Y, X
